
from enum import Enum

from wx import Colour
from wx import ColourDatabase


class MiniOglColorEnum(Enum):
    """
    The purpose of this enumeration is to keep wxPython types from getting into
    the user visible portions of the preferences dialog
    Custom colors came from:
        https://www.rapidtables.com/web/color/RGB_Color.html
    """

    BLACK             = 'Black'
    CORNFLOWER_BLUE   = 'Cornflower Blue'
    WHITE             = 'White'
    LIGHT_GREY        = 'Light Grey'
    DARK_GREY         = 'Dark Grey'
    SKY_BLUE          = 'Sky Blue'
    DIM_GREY          = 'Dim Grey'
    GREEN             = 'Green'
    MEDIUM_BLUE       = 'Medium Blue'
    MIDNIGHT_BLUE     = 'Midnight Blue'
    LIGHT_BLUE        = 'Light Blue'
    LIGHT_STEEL_BLUE  = 'Light Steel Blue'
    ALICE_BLUE        = 'Alice Blue'
    DARK_SLATE_BLUE   = 'Dark Slate Blue'
    MEDIUM_SLATE_BLUE = 'Medium Slate Blue'
    YELLOW            = 'Yellow'
    SALMON            = 'Salmon'
    GAINSBORO         = 'Gainsboro'
    LIGHT_YELLOW      = 'Light Yellow'
    MINT_CREAM        = 'Mint Cream'

    @staticmethod
    def toWxColor(colorEnum: 'MiniOglColorEnum') -> Colour:
        cdb: ColourDatabase = ColourDatabase()

        cdb.AddColour(MiniOglColorEnum.ALICE_BLUE.value,   Colour(240, 248, 255))
        cdb.AddColour(MiniOglColorEnum.GAINSBORO.value,    Colour(218, 218, 218))
        cdb.AddColour(MiniOglColorEnum.LIGHT_YELLOW.value, Colour(255, 255, 224))
        cdb.AddColour(MiniOglColorEnum.MINT_CREAM.value,   Colour(245, 255, 250))

        c: Colour = cdb.Find(colorEnum.value)
        if c.IsOk() is False:
            c = cdb.Find(MiniOglColorEnum.BLACK.value)
            print('Cannot find color use default')

        return c




from logging import Logger
from logging import getLogger

from miniogl.LinePoint import LinePoint
from miniogl.Shape import Shape


class AnchorPoint(LinePoint):
    """
    This is a point which begins or ends a line.
    It is often anchored to a parent shape, but that's not mandatory.
    """
    def __init__(self, x: int, y: int, parent: Shape | None):
        """

        Args:
            x: x position of the point
            y: y position of the point
            parent:
        """
        super().__init__(x, y, parent)

        self.apLogger: Logger = getLogger(__name__)

        self.apLogger.debug(f'AnchorPoint __init__  x: {x}, y: {y} parent: {parent}')
        self._protected:    bool = True  # protected by default
        self._stayInside:   bool = True
        self._stayOnBorder: bool = True
        self.draggable = False

    def SetStayInside(self, state: bool):
        """
        If True, the point will stay inside the bounds of its parent shape.
        Args:
            state:
        """
        self._stayInside = state

    def GetStayInside(self) -> bool:
        """

        Returns:  `True` if the point stays inside the bounds of its parent shape.
        """
        return self._stayInside

    def SetStayOnBorder(self, state: bool):
        """
        If True, the point will stay on the border of its parent shape.
        Args:
            state:
        """
        self._stayOnBorder = state

    def GetStayOnBorder(self) -> bool:
        """
        Return True if the point stays on the border of its parent shape.

        @return boolean
        """
        return self._stayOnBorder

    def SetPosition(self, x: int, y: int):
        """
        Change the position of the anchor point, if it's draggable.

        Args:
            x:  Abscissa of anchor point
            y:  Ordinate of anchor point
        """

        self.apLogger.debug(
            (
                f'x,y: ({x},{y}) '
                f'parent: {self._parent} '
                f'draggable: {self._draggable} '
                f'stayInside: {self._stayInside} '
                f'stayOnBorder: {self._stayOnBorder}')
        )
        if self._draggable:
            if self._parent is None:
                self._x = x
                self._y = y
            else:
                topLeftX, topLeftY = self._parent.GetTopLeft()
                width, height      = self._parent.GetSize()
                width  = abs(width) - 1
                height = abs(height) - 1
                self.apLogger.debug(f'topLeftX,topLeftY ({topLeftX},{topLeftY}) width,height ({width},{height})')

                from miniogl.LineShape import LineShape    # avoid circular import

                if isinstance(self._parent, LineShape):
                    x = topLeftX
                else:
                    if self._stayInside or self._stayOnBorder:
                        x = self.stayInside(topLeftX, width, x)
                        y = self.stayInside(topLeftY, height, y)
                        if self._stayOnBorder:
                            x, y = self.stickToBorder(topLeftX, topLeftY, width, height, x, y)

                self._x, self._y = self.ConvertCoordToRelative(x, y)

                self.apLogger.debug(f'Final Position: ({self._x}, {self._y})')

            if self.HasDiagramFrame():
                self.UpdateModel()

    def stayInside(self, low, length, value):
        """
        Return the nearest value in [low, low+length].

        Args:
            low:
            length:
            value:

        Returns: The nearest value
        """
        if value < low:
            value = low
        elif value > low + length:
            value = low + length
        return value

    def stickToBorder(self, ox, oy, width, height, x, y):
        """

        Args:
            ox:
            oy:
            width:
            height:
            x:
            y:

        Returns:  (x, y) on the square (ox, oy, ox+width, oy+height) by
        placing (x,y) on the nearest border.
        """
        left  = x - ox
        right = ox + width - x
        up    = y - oy
        down  = oy + height - y

        choice = {
            left: lambda xLeft, yLeft: (ox, y),
            right: lambda xRight, yRight: (ox + width, y),
            up: lambda xUp, yUp: (x, oy),
            down: lambda xDown, yDown: (x, oy + height),
        }
        lesser = min(left, right, up, down)
        self.apLogger.debug(f'lesser: {lesser}')
        return choice[lesser](x, y)

    def Detach(self):
        """
        Detach the line and all its line points, including src and dst.
        Also remove self from the parent.
        """
        LinePoint.Detach(self)
        parent = self.GetParent()
        if parent:
            parent.RemoveAnchor(self)

    def __str__(self) -> str:
        x, y = self.GetPosition()
        draggable: bool = self._draggable
        return f'AnchorPoint[({x},{y}) - {draggable=}]'

    def __repr__(self):
        return self.__str__()




from enum import Enum


class AttachmentSide(Enum):
    """
    Cardinal points, taken to correspond to the attachment points of the OglClass
    """
    NORTH = 0
    EAST  = 1
    SOUTH = 2
    WEST  = 3

    def __str__(self):
        return str(self.name)

    @staticmethod
    def toEnum(strValue: str) -> 'AttachmentSide':
        """
        Converts the input string to the attachment location
        Args:
            strValue:   A serialized string value

        Returns:  The visibility enumeration
        """
        canonicalStr: str = strValue.strip(' ')

        if canonicalStr == 'NORTH':
            return AttachmentSide.NORTH
        elif canonicalStr == 'EAST':
            return AttachmentSide.EAST
        elif canonicalStr == 'WEST':
            return AttachmentSide.WEST
        elif canonicalStr == 'SOUTH':
            return AttachmentSide.SOUTH
        else:
            print(f'Warning: did not recognize this attachment point: {canonicalStr}')
            return AttachmentSide.NORTH




from typing import List
from typing import cast

from logging import Logger
from logging import getLogger

from wx import CB_READONLY
from wx import EVT_CHECKBOX
from wx import EVT_COMBOBOX

from wx import CheckBox
from wx import ComboBox
from wx import CommandEvent
from wx import Window

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

from hasiicommon.ui.widgets.DimensionsControl import DimensionsControl

from ogl.OglDimensions import OglDimensions
from ogl.OglTextFontFamily import OglTextFontFamily

from ogl.preferences.OglPreferences import OglPreferences


class TextAttributesControl(SizedPanel):

    def __init__(self, parent: Window):

        self.logger:       Logger         = getLogger(__name__)
        self._preferences: OglPreferences = OglPreferences()
        super().__init__(parent)

        self._textDimensions: DimensionsControl = DimensionsControl(sizedPanel=self,
                                                                    displayText='Text Width/Height',
                                                                    valueChangedCallback=self._onTextDimensionsChanged,
                                                                    setControlsSize=False
                                                                    )
        self._boldText:         CheckBox = cast(CheckBox, None)
        self._italicizeText:    CheckBox = cast(CheckBox, None)
        self._fontSelector:     ComboBox = cast(ComboBox, None)
        self._fontSizeSelector: ComboBox = cast(ComboBox, None)

        self._createTextStylePanel(self)
        self._createFontAttributesPanel(self)
        self._setControlValues()
        self._bindControls()

    def _setControlValues(self):

        self._textDimensions.dimensions = self._preferences.textDimensions
        self._boldText.SetValue(self._preferences.textBold)
        self._italicizeText.SetValue(self._preferences.textItalicize)
        self._fontSizeSelector.SetValue(str(self._preferences.textFontSize))

    def _bindControls(self):

        self.Bind(EVT_CHECKBOX, self._onTextBoldValueChanged,      self._boldText)
        self.Bind(EVT_CHECKBOX, self._onTextItalicizeValueChanged, self._italicizeText)

        self.Bind(EVT_COMBOBOX, self._onFontSelectionChanged,     self._fontSelector)
        self.Bind(EVT_COMBOBOX, self._onFontSizeSelectionChanged, self._fontSizeSelector)

    def _createTextStylePanel(self, parent: SizedPanel):

        stylePanel: SizedPanel = SizedPanel(parent)
        stylePanel.SetSizerType('horizontal')
        stylePanel.SetSizerProps(expand=True, proportion=1)

        self._boldText      = CheckBox(parent=stylePanel, label='Bold Text')
        self._italicizeText = CheckBox(parent=stylePanel, label='Italicize Text')

    def _createFontAttributesPanel(self, parent: SizedPanel):

        fontChoices = []
        for fontName in OglTextFontFamily:
            fontChoices.append(fontName.value)

        fontPanel: SizedStaticBox = SizedStaticBox(parent, label='Font Family and Size')
        fontPanel.SetSizerType('horizontal')
        fontPanel.SetSizerProps(expand=True, proportion=1)

        self._fontSelector = ComboBox(fontPanel, choices=fontChoices, style=CB_READONLY)
        self._fontSelector.SetSizerProps(expand=True, proportion=1)

        fontSizes: List[str] = ['8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20']
        self._fontSizeSelector = ComboBox(fontPanel, choices=fontSizes, style=CB_READONLY)
        self._fontSizeSelector.SetSizerProps(expand=True, proportion=1)

    def _onTextDimensionsChanged(self, newValue: OglDimensions):
        self._preferences.textDimensions = newValue

    def _onTextBoldValueChanged(self, event: CommandEvent):

        val: bool = event.IsChecked()

        self._preferences.textBold = val

    def _onTextItalicizeValueChanged(self, event: CommandEvent):

        val: bool = event.IsChecked()
        self._preferences.textItalicize = val

    def _onFontSelectionChanged(self, event: CommandEvent):

        newFontName: str          = event.GetString()
        fontEnum:    OglTextFontFamily = OglTextFontFamily(newFontName)

        self._preferences.textFontFamily = fontEnum

    def _onFontSizeSelectionChanged(self, event: CommandEvent):

        newFontSize: str = event.GetString()

        self._preferences.textFontSize = int(newFontSize)




from dataclasses import dataclass
from typing import Tuple


@dataclass
class OglPosition:

    x: int = 0
    y: int = 0

    @classmethod
    def tupleToOglPosition(cls, position: Tuple[int, int]) -> 'OglPosition':
        """
        tuple[0] is the abscissa
        tuple[1] is the ordinate

        Args:
            position:  A position in Tuple format,

        Returns:  An OglPosition object
        """

        return OglPosition(x=position[0], y=position[1])



from typing import Callable
from typing import List
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from wx import EVT_TEXT
from wx import ID_ANY

from wx import CommandEvent
from wx import StaticText
from wx import TextCtrl
from wx import Window

from wx.lib.sized_controls import SizedPanel

from ogl.preferences.OglPreferences import OglPreferences


@dataclass
class NameData:
    textCtrl:     TextCtrl = cast(TextCtrl, None)
    label:        str      = ''
    initialValue: str = ''
    callback:     Callable = cast(Callable, None)


class DefaultNamesControl(SizedPanel):
    """
    A form for all the default object names
    """

    def __init__(self, parent: Window):

        self.logger:       Logger          = getLogger(__name__)
        self._preferences: OglPreferences = OglPreferences()
        super().__init__(parent)

        self.SetSizerType('form')
        #
        p: OglPreferences = self._preferences
        self._nameData: List[NameData] = [
            NameData(label='Interface Name: ', callback=self._onInterfaceNameChanged, initialValue=p.interfaceName),
            NameData(label='Use Case Name: ',  callback=self._onUseCaseNameChanged,   initialValue=p.useCaseName),
            NameData(label='Actor Name: ',     callback=self._onActorNameChanged,     initialValue=p.actorName),
            NameData(label='Method Name: ',    callback=self._onMethodNameChanged,    initialValue=p.methodName),
            NameData(label='Field Name: ',     callback=self._onFieldNameChanged,     initialValue=p.fieldName),
            NameData(label='Parameter Name: ', callback=self._onParameterNameChanged, initialValue=p.parameterName),
        ]
        for nd in self._nameData:
            nameData: NameData = cast(NameData, nd)
            StaticText(self, ID_ANY, nameData.label)
            nameData.textCtrl = TextCtrl(self, value=nameData.initialValue)
            nameData.textCtrl.SetSizerProps(expand=True)
            parent.Bind(EVT_TEXT, nameData.callback, nameData.textCtrl)

    def _onInterfaceNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.interfaceName = newValue

    def _onUseCaseNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.useCaseName = newValue

    def _onActorNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.actorName = newValue

    def _onMethodNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.methodName = newValue

    def _onFieldNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.fieldName = newValue

    def _onParameterNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.parameterName = newValue



from enum import Enum

from wx import PENSTYLE_CROSS_HATCH
from wx import PENSTYLE_DOT
from wx import PENSTYLE_DOT_DASH
from wx import PENSTYLE_HORIZONTAL_HATCH
from wx import PENSTYLE_LONG_DASH
from wx import PENSTYLE_SHORT_DASH
from wx import PENSTYLE_SOLID
from wx import PENSTYLE_VERTICAL_HATCH


class MiniOglPenStyle(Enum):

    SOLID                   = 'Solid'
    DOT                     = 'Dot'
    LONG_DASH               = 'Long Dash'
    SHORT_DASH              = 'Short Dash'
    DOT_DASH                = 'Dot Dash'
    USER_DASH               = 'User Dash'

    CROSS_HATCH             = 'Cross Hatch'
    HORIZONTAL_HATCH        = 'Horizontal Hatch'
    VERTICAL_HATCH          = 'Vertical Hatch'

    @staticmethod
    def toWxPenStyle(penStyleEnum: 'MiniOglPenStyle'):

        if penStyleEnum == MiniOglPenStyle.SOLID:
            wxPenStyle = PENSTYLE_SOLID

        elif penStyleEnum == MiniOglPenStyle.DOT:
            wxPenStyle = PENSTYLE_DOT

        elif penStyleEnum == MiniOglPenStyle.LONG_DASH:
            wxPenStyle = PENSTYLE_LONG_DASH

        elif penStyleEnum == MiniOglPenStyle.SHORT_DASH:
            wxPenStyle = PENSTYLE_SHORT_DASH

        elif penStyleEnum == MiniOglPenStyle.DOT_DASH:
            wxPenStyle = PENSTYLE_DOT_DASH

        elif penStyleEnum == MiniOglPenStyle.CROSS_HATCH:
            wxPenStyle = PENSTYLE_CROSS_HATCH

        elif penStyleEnum == MiniOglPenStyle.HORIZONTAL_HATCH:
            wxPenStyle = PENSTYLE_HORIZONTAL_HATCH

        elif penStyleEnum == MiniOglPenStyle.VERTICAL_HATCH:
            wxPenStyle = PENSTYLE_VERTICAL_HATCH

        else:
            wxPenStyle = PENSTYLE_SOLID

        return wxPenStyle



from logging import Logger
from logging import getLogger
from typing import cast

from hasiihelper.Dimensions import Dimensions

from hasiicommon.ui.widgets.DimensionsControl import DimensionsControl

from wx import CommandEvent
from wx import EVT_TEXT
from wx import SpinCtrl
from wx import Window

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

from ogl.OglDimensions import OglDimensions
from ogl.preferences.OglPreferences import OglPreferences


class SDAttributesControl(SizedPanel):

    def __init__(self, parent: Window):

        self.logger:       Logger          = getLogger(__name__)
        self._preferences: OglPreferences  = OglPreferences()
        super().__init__(parent)
        self.SetSizerType('horizontal')

        self._instancePosition:   SpinCtrl          = cast(SpinCtrl, None)
        self._instanceDimensions: DimensionsControl = cast(DimensionsControl, None)

        self._layoutControls(parentSizedPanel=self)
        self._setControlValues()

    def _layoutControls(self, parentSizedPanel: SizedPanel):
        self._layoutInstancePositionControl(parentSizedPanel=parentSizedPanel)

        self._instanceDimensions = DimensionsControl(sizedPanel=parentSizedPanel, displayText='Instance Width/Height',
                                                     minValue=100, maxValue=1000,
                                                     valueChangedCallback=self._noteDimensionsChanged,
                                                     setControlsSize=False)

    def _setControlValues(self):
        """
        """
        self._instancePosition.SetValue(self._preferences.instanceYPosition)
        width:  int = self._preferences.instanceDimensions.width
        height: int = self._preferences.instanceDimensions.height
        self._instanceDimensions.dimensions = Dimensions(width=width, height=height)

    def _layoutInstancePositionControl(self, parentSizedPanel: SizedPanel):

        instancePositionSSB: SizedStaticBox = SizedStaticBox(parentSizedPanel, label='Instance Position')
        instancePositionSSB.SetSizerProps(expand=True, proportion=1)
        self._instancePosition = SpinCtrl(parent=instancePositionSSB)

        # Bind to the text control;  Then we can type in or spin
        self.Bind(EVT_TEXT, self._onPositionsChanged, self._instancePosition)

    def _onPositionsChanged(self, event: CommandEvent):

        newValue: int = event.GetInt()
        self._preferences.instanceYPosition = newValue

    def _noteDimensionsChanged(self, newValue: Dimensions):
        self._preferences.instanceDimensions = OglDimensions(width=newValue.width, height=newValue.height)




from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from wx import DC
from wx import EVT_MENU
from wx import FONTFAMILY_SWISS
from wx import FONTSTYLE_NORMAL
from wx import FONTWEIGHT_BOLD

from wx import Font
from wx import ClientDC
from wx import Menu
from wx import CommandEvent
from wx import MenuItem
from wx import MouseEvent
from wx import Point
from wx import Brush
from wx import Colour

from miniogl.MiniOglColorEnum import MiniOglColorEnum
from miniogl.SelectAnchorPoint import SelectAnchorPoint

from pyutmodel.PyutDisplayParameters import PyutDisplayParameters
from pyutmodel.PyutMethod import PyutMethod
from pyutmodel.PyutObject import PyutObject
from pyutmodel.PyutClass import PyutClass
from pyutmodel.PyutStereotype import PyutStereotype

from ogl.OglObject import OglObject
from ogl.OglObject import DEFAULT_FONT_SIZE

from ogl.OglConstants import OglConstants
from ogl.OglUtils import OglUtils
from ogl.events.OglEvents import OglEventType

from ogl.preferences.OglPreferences import OglPreferences

# Menu IDs
[
    MENU_TOGGLE_STEREOTYPE,
    MENU_TOGGLE_FIELDS,
    MENU_TOGGLE_METHODS,
    MENU_TOGGLE_METHOD_PARAMETERS,
    MENU_FIT_FIELDS,
    MENU_CUT_SHAPE,
    MENU_IMPLEMENT_INTERFACE
]  = OglUtils.assignID(7)

MARGIN: int = 10


@dataclass
class ClickedOnSelectAnchorPointData:
    clicked:           bool             = False
    selectAnchorPoint: SelectAnchorPoint = cast(SelectAnchorPoint, None)


class OglClass(OglObject):
    """
    OGL object that represents a UML class in class diagrams.
    This class defines OGL objects that represents a class. You can just
    instantiate an OGL class and add it to the diagram, links, resizing,
    ... are managed by parent class `OglObject`.

    For more instructions about how to create an OGL object, please refer
    to the `OglObject` class.
    """
    def __init__(self, pyutClass: PyutClass | None, w: int = 0, h: int = 0):
        """

        Args:
            pyutClass:  a PyutClass object
            w:  Width of the shape
            h:  Height of the shape
        """
        if pyutClass is None:
            pyutObject = PyutClass()
        else:
            pyutObject = pyutClass

        width:  int = w
        height: int = h

        self._oglPreferences: OglPreferences = OglPreferences()

        # Use preferences to get initial size if not specified
        # Note: auto_resize_shape_on_edit must be False for this size to actually stick
        if w == 0:
            width = self._oglPreferences.classDimensions.width
        if h == 0:
            height = self._oglPreferences.classDimensions.height

        super().__init__(pyutObject, width=width, height=height)

        self._nameFont: Font = Font(DEFAULT_FONT_SIZE, FONTFAMILY_SWISS, FONTSTYLE_NORMAL, FONTWEIGHT_BOLD)
        oglTextColor:   MiniOglColorEnum = self._oglPreferences.classTextColor
        self._textColor: Colour = Colour(MiniOglColorEnum.toWxColor(oglTextColor))
        oglBackgroundColor: MiniOglColorEnum = self._oglPreferences.classBackgroundColor
        backgroundColor: Colour = Colour(MiniOglColorEnum.toWxColor(oglBackgroundColor))

        self.SetBrush(Brush(backgroundColor))

        self.logger:    Logger = getLogger(__name__)

    def handleSelectAnchorPointSelection(self, event: MouseEvent):
        """
        May be called (inexcusably bad form) by the selection anchor point left down handler
        by using its parent protected attribute

        Args:
            event:
        """
        self.logger.info(f'OnLeftDown: {event.GetPosition()}')
        # noinspection PyPropertyAccess
        clickPoint: Point = event.Position
        selectData: ClickedOnSelectAnchorPointData = self._didWeClickOnSelectAnchorPoint(clickPoint=clickPoint)
        if selectData.clicked is True:
            self.eventEngine.sendEvent(OglEventType.CreateLollipopInterface, implementor=self, attachmentPoint=selectData.selectAnchorPoint)

    def GetTextWidth(self, dc, text):
        width = dc.GetTextExtent(text)[0]
        return width

    def GetTextHeight(self, dc, text):
        height = dc.GetTextExtent(text)[1]
        return height

    def Draw(self, dc, withChildren=False):
        """
        Paint handler, draws the content of the shape.

        WARNING : Every change here must be reported in autoResize pyutMethod

        Args:
            dc: device context to draw to
            withChildren:
        """

        pyutObject: PyutClass = cast(PyutClass, self.pyutObject)

        # Draw rectangle shape
        OglObject.Draw(self, dc)

        # drawing is restricted in the specified region of the device
        w, h = self._width, self._height
        x, y = self.GetPosition()           # Get position
        dc.SetClippingRegion(x, y, w, h)

        # Draw header
        (headerX, headerY, headerW, headerH) = self._drawClassHeader(dc, True)
        y = headerY + headerH

        if pyutObject.showFields is True:
            # Draw line
            dc.DrawLine(x, y, x + w, y)

            # Draw fields
            (fieldsX, fieldsY, fieldsW, fieldsH) = self._drawClassFields(dc, True, initialY=y)
            y = fieldsY + fieldsH
        # Draw line
        dc.DrawLine(x, y, x + w, y)
        #
        # Method needs to be called even though returned values not used  -- TODO look at refactoring
        #
        if pyutObject.showMethods is True:
            (methodsX, methodsY, methodsW, methodsH) = self._drawClassMethods(dc, True, initialY=y, calcWidth=True)
            # noinspection PyUnusedLocal
            y = methodsY + methodsH

        dc.DestroyClippingRegion()

    def autoResize(self):
        """
        Auto-resize the class

        WARNING : Every change here must be reported in DRAW pyutMethod
        """
        # Init
        pyutObject: PyutClass = cast(PyutClass, self.pyutObject)
        umlFrame = self.GetDiagram().GetPanel()
        dc = ClientDC(umlFrame)

        # Get header size
        (headerX, headerY, headerW, headerH) = self._drawClassHeader(dc, False, calcWidth=True)
        y = headerY + headerH

        # Get fields size
        if pyutObject.showFields is True:
            (fieldsX, fieldsY, fieldsW, fieldsH) = self._drawClassFields(dc, False, initialY=y, calcWidth=True)
            y = fieldsY + fieldsH
        else:
            fieldsW, fieldsH = 0, 0

        # Get methods size
        if pyutObject.showMethods is True:
            (methodX, methodY, methodW, methodH) = self._drawClassMethods(dc, True, initialY=y, calcWidth=True)
            y = methodY + methodH
        else:
            methodW, methodH = 0, 0

        w = max(headerW, fieldsW, methodW)
        h = y - headerY
        w += 2 * MARGIN
        self.SetSize(w, h)

        # to automatically replace the sizer objects at a correct place
        if self.selected is True:
            self.selected = False
            self.selected = True

    def OnRightDown(self, event: MouseEvent):
        """
        Callback for right clicks
        """
        pyutObject: PyutClass = cast(PyutClass, self.pyutObject)
        menu:       Menu      = Menu()

        menu.Append(MENU_TOGGLE_STEREOTYPE, "Toggle stereotype display", "Set stereotype display on or off", True)
        item = menu.FindItemById(MENU_TOGGLE_STEREOTYPE)
        item.Check(pyutObject.displayStereoType)

        menu.Append(MENU_TOGGLE_FIELDS, "Toggle fields display", "Set fields display on or off", True)
        item = menu.FindItemById(MENU_TOGGLE_FIELDS)
        item.Check(pyutObject.showFields)

        menu.Append(MENU_TOGGLE_METHODS, "Toggle methods display", "Set methods display on or off ", True)
        item = menu.FindItemById(MENU_TOGGLE_METHODS)
        item.Check(pyutObject.showMethods)

        menu.Append(MENU_TOGGLE_METHOD_PARAMETERS, "Toggle parameter display", "Set parameter display on or off", True)

        itemToggleParameters: MenuItem              = menu.FindItemById(MENU_TOGGLE_METHOD_PARAMETERS)
        displayParameters:    PyutDisplayParameters = self.pyutObject.displayParameters

        self._initializeTriStateDisplayParametersMenuItem(displayParameters, itemToggleParameters)

        menu.Append(MENU_FIT_FIELDS, "Fit Fields", "Fit to see all class fields")
        menu.Append(MENU_CUT_SHAPE,  "Cut shape",  "Cut this shape")

        menu.Append(MENU_IMPLEMENT_INTERFACE, 'Implement Interface', 'Use Existing interface or create new one')

        frame = self._diagram.GetPanel()

        # Callback
        menu.Bind(EVT_MENU, self.OnMenuClick, id=MENU_TOGGLE_STEREOTYPE)
        menu.Bind(EVT_MENU, self.OnMenuClick, id=MENU_TOGGLE_FIELDS)
        menu.Bind(EVT_MENU, self.OnMenuClick, id=MENU_TOGGLE_METHODS)
        menu.Bind(EVT_MENU, self.OnMenuClick, id=MENU_FIT_FIELDS)
        menu.Bind(EVT_MENU, self.OnMenuClick, id=MENU_CUT_SHAPE)
        menu.Bind(EVT_MENU, self.OnMenuClick, id=MENU_IMPLEMENT_INTERFACE)
        menu.Bind(EVT_MENU, self.onDisplayParametersClick, id=MENU_TOGGLE_METHOD_PARAMETERS)

        x: int = event.GetX()
        y: int = event.GetY()
        self.logger.debug(f'OglClass - x,y: {x},{y}')
        frame.PopupMenu(menu, x, y)

    def OnMenuClick(self, event: CommandEvent):
        """
        Callback for popup menu on class

        Args:
            event:
        """
        pyutObject:   PyutClass = cast(PyutClass, self.pyutObject)
        eventId:      int       = event.GetId()

        if eventId == MENU_TOGGLE_STEREOTYPE:
            pyutObject.setShowStereotype(not pyutObject.displayStereoType)
            self.autoResize()
        elif eventId == MENU_TOGGLE_METHODS:
            pyutObject.showMethods = not pyutObject.showMethods     # flip it!!  too cute
            self.autoResize()
        elif eventId == MENU_TOGGLE_FIELDS:
            pyutObject.showFields = not pyutObject.showFields       # flip it!! too cute
            self.autoResize()
        elif eventId == MENU_FIT_FIELDS:
            self.autoResize()
        elif eventId == MENU_CUT_SHAPE:
            self.eventEngine.sendEvent(OglEventType.CutOglClass, shapeToCut=self)
        elif eventId == MENU_IMPLEMENT_INTERFACE:
            self.eventEngine.sendEvent(OglEventType.RequestLollipopLocation, requestShape=self)
        else:
            event.Skip()

    # noinspection PyUnusedLocal
    def onDisplayParametersClick(self, event: CommandEvent):
        """
        This menu item has its own handler because this option is tri-state

        Unspecified --> Display  --> Do Not Display ---|
            ^------------------------------------------|

        Args:
            event:
        """
        pyutClass:         PyutClass             = cast(PyutClass, self.pyutObject)
        displayParameters: PyutDisplayParameters = pyutClass.displayParameters
        self.logger.debug(f'Current: {displayParameters=}')

        if displayParameters == PyutDisplayParameters.UNSPECIFIED:
            pyutClass.displayParameters = PyutDisplayParameters.DISPLAY
        elif displayParameters == PyutDisplayParameters.DISPLAY:
            pyutClass.displayParameters = PyutDisplayParameters.DO_NOT_DISPLAY
        elif displayParameters == PyutDisplayParameters.DO_NOT_DISPLAY:
            pyutClass.displayParameters = PyutDisplayParameters.UNSPECIFIED
        else:
            assert False, 'Unknown display type'
        self.logger.warning(f'New: {pyutClass.displayParameters=}')

    def _didWeClickOnSelectAnchorPoint(self, clickPoint: Point) -> ClickedOnSelectAnchorPointData:
        """

        Args:
            clickPoint:

        Returns:  Data class with relevant information
        """
        from miniogl.Shape import Shape

        selectData: ClickedOnSelectAnchorPointData = ClickedOnSelectAnchorPointData(clicked=False)
        anchors = self.GetAnchors()
        for shape in anchors:
            child: Shape = cast(Shape, shape)
            if isinstance(child, SelectAnchorPoint):
                selectAnchorPoint: SelectAnchorPoint = cast(SelectAnchorPoint, child)
                x, y = clickPoint.Get()
                if selectAnchorPoint.Inside(x=x, y=y):
                    selectData.selectAnchorPoint = child
                    selectData.clicked           = True
                    break

        return selectData

    def _isSameName(self, other) -> bool:

        ans: bool = False
        selfPyutObj:  PyutObject = self.pyutObject
        otherPyutObj: PyutObject = other.pyutObject

        if selfPyutObj.name == otherPyutObj.name:
            ans = True
        return ans

    def _isSameId(self, other):

        ans: bool = False
        if self.id == other.id:
            ans = True
        return ans

    def _initializeTriStateDisplayParametersMenuItem(self, displayParameters: PyutDisplayParameters, itemToggleParameters: MenuItem):

        if displayParameters == PyutDisplayParameters.UNSPECIFIED:
            itemToggleParameters.SetBitmap(OglConstants.unspecifiedDisplayMethodsIcon())
        elif displayParameters == PyutDisplayParameters.DISPLAY:
            itemToggleParameters.SetBitmap(OglConstants.displayMethodsIcon())
        elif displayParameters == PyutDisplayParameters.DO_NOT_DISPLAY:
            itemToggleParameters.SetBitmap(OglConstants.doNotDisplayMethodsIcon())
        else:
            assert False, 'Unknown display type'

    def _drawClassHeader(self, dc: DC, draw: bool = False, initialX=None, initialY=None, calcWidth: bool = False):
        """
        Calculate the class header position and size and display it if
        a draw is True

        Args:
            dc:
            draw:
            initialX:
            initialY:
            calcWidth:

        Returns:    tuple (x, y, w, h) = position and size of the header
        """
        # Init
        dc.SetFont(self._defaultFont)
        dc.SetTextForeground(self._textColor)
        # pyutObject = self.getPyutObject()
        x, y = self.GetPosition()
        if initialX is not None:
            x = initialX
        if initialY is not None:
            y = initialY
        w = self._width
        h = 0
        if calcWidth:
            w = 0

        # define space between text and line
        lth = dc.GetTextExtent("*")[1] // 2

        # from where begin the text
        h += lth

        # draw a pyutClass name
        name: str = self.pyutObject.name
        dc.SetFont(self._nameFont)
        nameWidth: int = self.GetTextWidth(dc, name)
        if draw:
            dc.DrawText(name, x + (w - nameWidth) // 2, y + h)
        if calcWidth:
            w = max(nameWidth, w)
        dc.SetFont(self._defaultFont)
        h += self.GetTextHeight(dc, str(name))
        h += lth
        #
        # Draw the stereotype value
        #
        pyutClass: PyutClass = self.pyutObject
        stereo: PyutStereotype = pyutClass.stereotype

        if pyutClass.displayStereoType is True and stereo is not None and stereo != PyutStereotype.NO_STEREOTYPE:
            stereoTypeValue: str = f'<<{stereo.value}>>'
        else:
            stereoTypeValue = ''

        stereoTypeValueWidth = self.GetTextWidth(dc, stereoTypeValue)
        if draw:
            dc.DrawText(stereoTypeValue, x + (w - stereoTypeValueWidth) // 2, y + h)
        if calcWidth:
            w = max(stereoTypeValueWidth, w)
        h += self.GetTextHeight(dc, str(stereoTypeValue))
        h += lth

        # Return sizes
        return x, y, w, h

    def _drawClassFields(self, dc, draw=False, initialX=None, initialY=None, calcWidth=False):
        """
        Calculate the class fields position and size and display it if
        a draw is True

        Args:
            dc:
            draw:
            initialX:
            initialY:
            calcWidth:

        Returns:    tuple : (x, y, w, h) = position and size of the field
        """
        # Init
        dc.SetFont(self._defaultFont)
        dc.SetTextForeground(self._textColor)

        x, y = self.GetPosition()
        if initialX is not None:
            x = initialX
        if initialY is not None:
            y = initialY
        w = self._width
        h = 0
        if calcWidth:
            w = 0

        # define space between text and line
        lth: int = dc.GetTextExtent("*")[1] // 2

        # Add space
        pyutClass: PyutClass = cast(PyutClass, self.pyutObject)
        if len(pyutClass.fields) > 0:
            h += lth

        # draw pyutClass fields
        if pyutClass.showFields is True:
            for field in pyutClass.fields:
                if draw:
                    dc.DrawText(str(field), x + MARGIN, y + h)
                if calcWidth:
                    w = max(w, self.GetTextWidth(dc, str(field)))

                h += self.GetTextHeight(dc, str(field))

        # Add space
        if len(pyutClass.fields) > 0:
            h += lth

        # Return sizes
        return x, y, w, h

    def _drawClassMethods(self, dc, draw=True, initialX=None, initialY=None, calcWidth=False) -> Tuple[int, int, int, int]:
        """
        Calculate the class methods position and size and display it if
        a draw is True

        Args:
            dc:
            draw:
            initialX:
            initialY:
            calcWidth:

        Returns:    tuple : (x, y, w, h) = position and size of the methods
        """

        dc.SetFont(self._defaultFont)
        dc.SetTextForeground(self._textColor)

        x, y = self.GetPosition()
        if initialX is not None:
            x = initialX
        if initialY is not None:
            y = initialY
        w = self._width
        h = 0
        if calcWidth:
            w = 0

        # define space between text and line
        lth = dc.GetTextExtent("*")[1] // 2

        # Add space
        pyutClass: PyutClass = cast(PyutClass, self.pyutObject)
        if len(pyutClass.methods) > 0:
            h += lth

        # draw pyutClass methods
        self.logger.debug(f"showMethods => {pyutClass.showMethods}")
        if pyutClass.showMethods is True:
            for method in pyutClass.methods:
                if draw is True:
                    self.__drawMethodSignature(dc, method, pyutClass, x, y, h)

                if calcWidth:
                    w = max(w, self.GetTextWidth(dc, str(method)))

                h += self.GetTextHeight(dc, str(method))

        # Add space
        if len(pyutClass.methods) > 0:
            h += lth

        # Return sizes
        return x, y, w, h

    def __drawMethodSignature(self, dc: DC, pyutMethod: PyutMethod, pyutClass: PyutClass, x: int, y: int, h: int):
        """
        If preference is not set at individual class level defer to global; Otherwise,
        respect the class level preference

        Args:
            dc:
            pyutMethod:
            pyutClass:
            x:
            y:
            h:
        """
        self.logger.debug(f'{pyutClass.displayParameters=} - {self._oglPreferences.showParameters=}')
        dc.SetTextForeground(self._textColor)
        if pyutClass.displayParameters == PyutDisplayParameters.UNSPECIFIED:
            if self._oglPreferences.showParameters is True:
                dc.DrawText(pyutMethod.methodWithParameters(), x + MARGIN, y + h)
            else:
                dc.DrawText(pyutMethod.methodWithoutParameters(), x + MARGIN, y + h)
        elif pyutClass.displayParameters == PyutDisplayParameters.DISPLAY:
            dc.DrawText(pyutMethod.methodWithParameters(), x + MARGIN, y + h)
        elif pyutClass.displayParameters == PyutDisplayParameters.DO_NOT_DISPLAY:
            dc.DrawText(pyutMethod.methodWithoutParameters(), x + MARGIN, y + h)
        else:
            assert False, 'Internal error unknown pyutMethod parameter display type'

    def __repr__(self) -> str:
        selfName: str = self.pyutObject.name
        modelId:  int = self.pyutObject.id
        return f'OglClass.{selfName} modelId: {modelId}'

    def __eq__(self, other) -> bool:

        if isinstance(other, OglClass):
            if self._isSameName(other) is True and self._isSameId(other) is True:
                return True
            else:
                return False
        else:
            return False

    def __hash__(self):

        selfPyutObj:  PyutObject = self.pyutObject

        return hash(selfPyutObj.name) + hash(self.id)




from typing import cast

from logging import Logger
from logging import getLogger

from deprecated import deprecated
from pyutmodel.PyutSDInstance import PyutSDInstance
from wx import BLACK_DASHED_PEN
from wx import PENSTYLE_LONG_DASH

from wx import Colour
from wx import Pen

from miniogl.AnchorPoint import AnchorPoint
from miniogl.LineShape import LineShape
from miniogl.RectangleShape import RectangleShape

from ogl.OglObject import OglObject
from ogl.preferences.OglPreferences import OglPreferences

from ogl.sd.OglInstanceName import OglInstanceName


class OglSDInstance(OglObject):
    """
    Class Diagram Instance
    This class is an OGL object for class diagram instance (vertical line)
    """

    def __init__(self, pyutObject: PyutSDInstance):
        """
        """
        self.logger:       Logger         = getLogger(__name__)

        prefs: OglPreferences = OglPreferences()    # need our own since superclass constructs it
        super().__init__(pyutObject, prefs.instanceDimensions.width, prefs.instanceDimensions.height)

        self._instanceYPosition: int = self._prefs.instanceYPosition  # User super class version

        self.draggable = True
        self.SetVisible(True)
        self.SetPen(Pen(Colour(200, 200, 255), 1, PENSTYLE_LONG_DASH))
        self.SetPosition(self.GetPosition()[0], self._instanceYPosition)

        self._lifeLine:     LineShape       = self._createLifeLine()
        self._instance:     RectangleShape  = self._createInstance()
        self._instanceName: OglInstanceName = self._createInstanceName(pyutSDInstance=pyutObject, instanceBox=self._instance)
        # TODO : set instance box size to the size of the text by invoking self._instanceBoxText.setSize()

    @deprecated(reason="Use the property '.lifeline'")
    def getLifeLineShape(self):
        """
        Used by OGLSDMessage to use it as parent

        Returns: The lifeline object
        """
        return self._lifeLine

    @property
    def lifeline(self) -> LineShape:
        """
        Used by OGLSDMessage to use it as parent and

        Returns: The lifeline object
        """
        return self._lifeLine

    @property
    def instance(self) -> RectangleShape:
        return self._instance

    @property
    def instanceName(self) -> OglInstanceName:
        return self._instanceName

    def OnInstanceBoxResize(self, sizer, width: int, height: int):
        """
        Resize the instance box, so all instance

        Args:
            sizer:
            width:
            height:

        """
        """

        @param double x, y : position of the sizer
        """
        RectangleShape.Resize(self._instance, sizer, width, height)
        size = self._instance.GetSize()
        self.SetSize(size[0], self.GetSize()[1])

    def Resize(self, sizer, width: int, height: int):
        """
        Resize the rectangle according to the new position of the sizer.

        Args:
            sizer:
            width:
            height:
        """
        OglObject.Resize(self, sizer, width, height)

    def SetSize(self, width: int, height: int):
        """
        """
        OglObject.SetSize(self, width, height)
        # Set lifeline
        (myX, myY) = self.GetPosition()
        (w, h) = self.GetSize()
        lineDst = self._lifeLine.destinationAnchor
        lineSrc = self._lifeLine.sourceAnchor
        lineSrc.draggable = True
        lineDst.draggable = True
        lineSrc.SetPosition(w // 2 + myX, 0 + myY)
        lineDst.SetPosition(w // 2 + myX, height + myY)
        lineSrc.draggable = False
        lineDst.draggable = False

        from ogl.sd.OglSDMessage import OglSDMessage
        # Update all OglSDMessage positions
        for link in self._oglLinks:
            try:
                oglSDMessage: OglSDMessage = cast(OglSDMessage, link)
                oglSDMessage.updatePositions()
            except (ValueError, Exception) as e:
                self.logger.error(f'Link update position error: {e}')
        # Set TextBox
        RectangleShape.SetSize(self._instance, width, self._instance.GetSize()[1])

    def SetPosition(self, x: int, y: int):
        """ 
        Debug
        """
        y = self._instanceYPosition
        OglObject.SetPosition(self, x, y)

    def Draw(self, dc, withChildren=True):
        """
        Draw override
        Args:
            dc:
            withChildren:  defaulted to True because of the child shapes
        """
        # Update labels
        self._instanceName.text = self._pyutObject.instanceName
        # Call parent's Draw method
        if self.selected is True:
            self.SetVisible(True)
            self.SetPen(Pen(Colour(200, 200, 255), 1, PENSTYLE_LONG_DASH))

        super().Draw(dc=dc, withChildren=withChildren)

    def OnLeftUp(self, event):
        """
        Callback for left clicks.
        """
        self.SetPosition(self.GetPosition()[0], self._instanceYPosition)

    def _createLifeLine(self) -> LineShape:
        """

        Returns:  The lifeline
        """
        width:   int = self._prefs.instanceDimensions.width
        height: int  = self._prefs.instanceDimensions.height
        (srcX, srcY, dstX, dstY) = (width // 2, 0,
                                    width // 2, height
                                    )

        (src, dst) = (AnchorPoint(srcX, srcY, self), AnchorPoint(dstX, dstY, self))
        for el in [src, dst]:
            el.SetVisible(False)
            el.draggable = False

        lifeLineShape: LineShape = LineShape(src, dst)

        lifeLineShape.SetParent(self)
        lifeLineShape.SetDrawArrow(False)
        lifeLineShape.draggable = True
        lifeLineShape.SetPen(BLACK_DASHED_PEN)
        lifeLineShape.SetVisible(True)

        self.AppendChild(lifeLineShape)

        return lifeLineShape

    def _createInstance(self) -> RectangleShape:
        """

        Returns:  The instance box
        """
        instanceBox: RectangleShape = RectangleShape(0, 0, 100, 50)

        instanceBox.draggable = False
        instanceBox.Resize = self.OnInstanceBoxResize   # type: ignore
        instanceBox.SetResizable(True)
        instanceBox.SetParent(self)

        self.AppendChild(instanceBox)

        return instanceBox

    def _createInstanceName(self, pyutSDInstance: PyutSDInstance, instanceBox: RectangleShape) -> OglInstanceName:
        """

        Returns:  An OglInstanceName
        """
        text: str = self._pyutObject.instanceName

        oglInstanceName: OglInstanceName = OglInstanceName(pyutSDInstance, 0, 20, text, instanceBox)

        self.AppendChild(oglInstanceName)

        return oglInstanceName

    def __str__(self) -> str:
        instanceName: str = self._pyutObject.instanceName
        return f'OglSDInstance[{self._id=} {instanceName=}]'

    def __repr__(self) -> str:
        return self.__str__()




from typing import List
from typing import cast

from logging import Logger
from logging import getLogger

from wx import CB_READONLY
from wx import CommandEvent
from wx import EVT_COMBOBOX
from wx import ID_ANY

from wx import ComboBox
from wx import StaticText
from wx import Window

from wx.lib.sized_controls import SizedPanel


from ogl.preferences.OglPreferences import OglPreferences


class AssociationAttributesControl(SizedPanel):

    def __init__(self, parent: Window):

        self.logger:       Logger         = getLogger(__name__)
        self._preferences: OglPreferences = OglPreferences()
        super().__init__(parent)
        self.SetSizerType('horizontal')

        self._textFontSize: ComboBox = cast(ComboBox, None)
        self._diamondSize:  ComboBox = cast(ComboBox, None)

        self._layoutControls(parentPanel=self)
        self._setControlValues()

        self.Bind(EVT_COMBOBOX, self._onTextFontSizedChanged, self._textFontSize)
        self.Bind(EVT_COMBOBOX, self._onDiamondSizeChanged,   self._diamondSize)

    def _layoutControls(self, parentPanel: SizedPanel):

        fontSizes:    List[str] = ['8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20']
        diamondSizes: List[str] = ['6', '7', '8', '10', '11', '12', '13', '14', '15']

        formPanel: SizedPanel = SizedPanel(parentPanel)
        formPanel.SetSizerType('form')

        # First Line
        StaticText(formPanel, ID_ANY, 'Font Size')
        self._textFontSize = ComboBox(formPanel, choices=fontSizes, style=CB_READONLY)

        # Second Line
        StaticText(formPanel, ID_ANY, 'Diamond Size')
        self._diamondSize = ComboBox(formPanel, choices=diamondSizes, style=CB_READONLY)

    def _setControlValues(self):
        self._textFontSize.SetValue(str(self._preferences.associationTextFontSize))
        self._diamondSize.SetValue(str(self._preferences.associationDiamondSize))

    def _onTextFontSizedChanged(self, event: CommandEvent):
        newFontSize: str = event.GetString()
        self._preferences.associationTextFontSize = int(newFontSize)

    def _onDiamondSizeChanged(self, event: CommandEvent):
        newDiamondSize: str = event.GetString()
        self._preferences.associationDiamondSize = int(newDiamondSize)




# ----------------------------------------------------------------------
# This file was generated by /Users/humberto.a.sanchez.ii/PycharmProjects/PyUt/venv-pyenv-3.9.0/bin/img2py
#
from wx.lib.embeddedimage import PyEmbeddedImage

embeddedImage: PyEmbeddedImage = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAK3RFWHRDcmVhdGlvbiBUaW1l'
    b'AEZyIDE1IE5vdiAyMDAyIDEwOjA4OjI5ICswMTAw9aJixgAAAAd0SU1FB9MDBgs0KlZxNO0A'
    b'AAAJcEhZcwAACxIAAAsSAdLdfvwAAAAEZ0FNQQAAsY8L/GEFAAACuUlEQVR42qWTXUhTYRjH'
    b'/2dnm57NzQ+UNVOMyj5ofVCRdVGQRhCY+UGWFnhlSBB0320hdBV1EZE3ERFCV0oqkaWBWUim'
    b'lTE3pvm1jmdnbmfz7Jyz89W7BYMSr3xenvfifd7nx/P+3+cBtmjU/wftM7vKRA43XfnF9XkW'
    b'525F10w2GQ0sCmyfXGU+XqwRYpsCuiKHW5wRT8/JytqiwnwX+PQC4hIPUUphOcZjdP4nP+fi'
    b'OhK12sAGQCd38HKV7Ht5vrKJDhnTCCSm8PTtZzgsDpw7sA2MaYepGRiYC6p+KtKIBgzkAK0z'
    b'+7wew+u/4ut0B7RxsMoSHrx+h3BzHLSVwpHBahzzOsHADt0w8Hz6W0Rk1L1oQ8ySAayFk7dO'
    b'ba9zL8jfsSLOZ5OXmqPZ5IxNXQjiV4JGTE1CVtM4WuwtwyK6MrEswJFX1FDKlGBGmMDDwRGw'
    b'LQLsVit004BJPGPDZ75Ao8rBqwk4GTug42IOwFic1cvJOTwb+QruahIUDVz71ABeYSGbChon'
    b'TmchvceHsd97FpKmII+i9+QAkq5hMuzHpYrrGOM+omO8DaNr/SiwuQmcwYo8hdvBVoytv8EJ'
    b'Tw1UE0TQv+JbM1tY4AJ1O3y+D+oQukP9qHSVoBpeUn6WjyJbIcKrU7jL3oCqGohFRKiSHsgB'
    b'Znm2bynG+crtbhg2Cyw6YJAvy6xsmTQNRSaCmjSEtTSSKRWmir7cE5Ri7dGr0KSgk/spRcG6'
    b'JJHLFNzWgiyAJqLImkoS0xDiClhOjEDDk2wss+m9WI+3i8FoRG6pKHRZRKKJTlrEr/sxqbxH'
    b'KDILnpcQJR7+LaqKTWvFPfzYOAsv0FSQsPUc8npKHPlWpNIyaeM0FIlCMqFglZM4jdE7cAdD'
    b'mw4T7qMUImkSE/WMnd6paxTSKS2INHkzQ8ruRnyrE/yP/QEVnEpqw3XnNAAAAABJRU5ErkJg'
    b'gg==')




from logging import Logger
from logging import getLogger

from deprecated import deprecated

from wx import BLACK
from wx import RED
from wx import Size
from wx import WHITE
from wx import PENSTYLE_SOLID
from wx import PENSTYLE_DOT

from wx import Font
from wx import Colour
from wx import MouseEvent
from wx import Pen

from wx import DC
from wx import MemoryDC

from miniogl.Shape import Shape
from miniogl.RectangleShape import RectangleShape
from miniogl.TextShapeModel import TextShapeModel


TEXT_Y_MARGIN: int = 2
TEXT_X_MARGIN: int = 3

DEFAULT_WIDTH:  int = 100
DEFAULT_HEIGHT: int = 24

TEXT_HEIGHT_ADJUSTMENT: int = 12
TEXT_WIDTH_ADJUSTMENT:  int = 24


class TextShape(RectangleShape):

    clsLogger: Logger = getLogger(__name__)
    """
    A text shape that can be attached to another shape standalone).
    """
    def __init__(self, x: int, y: int, text: str, parent=None, font: Font = None):
        """

        Args:
            x:          x position of the point
            y:          y position of the point
            text:       the text that the shape displays
            parent:     parent shape
            font:       Font to use
        """
        super().__init__(x, y, width=DEFAULT_WIDTH, height=DEFAULT_HEIGHT, parent=parent)

        self.text: str = text

        self._drawFrame: bool = False
        self._resizable: bool = False

        self._textColor:            Colour = BLACK
        self._textBackgroundColor:  Colour = WHITE

        self._redColor:    Colour = RED
        self._font:        Font   = font
        self._selectedPen: Pen = Pen(colour=RED, width=1, style=PENSTYLE_DOT)

        self._model: TextShapeModel = TextShapeModel(self)

    def Attach(self, diagram):
        """
        Do not use this method, use Diagram.AddShape instead !!!
        Attach the shape to a diagram.
        When you create a new shape, you must attach it to a diagram before
        you can see it. This method is used internally by Diagram.AddShape.

        Args:
            diagram
        """
        # RectangleShape.Attach(self, diagram)
        super().Attach(diagram)
        self._textBackgroundColor = self._diagram.GetPanel().GetBackgroundColour()

    @property
    def text(self) -> str:
        """
        Returns:  The text that the text shape displays
        """
        return self._text

    @text.setter
    def text(self, newValue: str):
        """
        Set the text that the shape displays

        Args:
              newValue
        """
        self._text = newValue

    @property
    def color(self) -> Colour:
        """
        Returns The text color
        """
        return self._textColor

    @color.setter
    def color(self, color: Colour):
        """
        Sets the color of the text.

        Args:
             color
        """
        self._textColor = color

    @property
    def font(self) -> Font:
        """
        Returns:  The text shape's private font
        """
        return self._font

    @deprecated(reason='Use the .text property')
    def GetText(self) -> str:
        """
        Returns:  The text that the text shape displays
        """
        return self._text

    @deprecated(reason='Use the .text property')
    def SetText(self, text: str):
        """
        Set the text that the shape displays

        Args:
              text
        """
        self._text = text
        self._width, self._height = MemoryDC().GetTextExtent(text)

    def SetTextBackground(self, color: Colour):
        """
        Set the text background color.

        Args:
             color
        """
        self._textBackgroundColor = color

    def GetTextBackground(self) -> Colour:
        """
        Get the text background color.

        Returns:
             the text background color
        """
        return self._textBackgroundColor

    def Draw(self, dc: DC, withChildren: bool = True):
        """
        Draw the text on the dc.

        Args:
            dc
            withChildren
        """
        if self._visible:
            super().Draw(dc=dc, withChildren=False)
            if self._selected:
                dc.SetPen(self._selectedPen)
                dc.SetTextForeground(self._redColor)
                self.DrawBorder(dc=dc)
            else:
                dc.SetTextForeground(self._textColor)

            self._computeTextSize(dc=dc)
            self._drawText(dc)

            if withChildren:
                self.DrawChildren(dc)

    def DrawBorder(self, dc: DC):
        """
        Draw the border of the shape, for fast rendering.

        Args:
            dc
        """
        if self._selected:
            RectangleShape.DrawBorder(self, dc)
        else:
            Shape.DrawBorder(self, dc)

    def UpdateFromModel(self):
        """
        Updates the shape position and size from the model in light of a
        change of state of the diagram frame.  Here it is only for the zoom
        """

        # change the position and size of the shape from the model
        # RectangleShape.UpdateFromModel(self)
        super().UpdateFromModel()
        # get the diagram frame ratio between the shape and the model
        ratio = self.GetDiagram().GetPanel().GetCurrentZoom()

        fontSize = round(self.GetModel().GetFontSize() * ratio)
        TextShape.clsLogger.debug(f'UpdateFromModel - ratio: {ratio}')

        # set the new font size
        if self._font is not None:
            self._font.SetPointSize(fontSize)

    def UpdateModel(self):
        """
        Updates the model when the shape (view) is displaced or resized.
        """
        # change the coordinates and size of model
        # RectangleShape.UpdateModel(self)
        super().UpdateModel()

        # get the ratio between the model and the shape (view) from
        # the diagram frame where the shape is displayed.
        ratio = self.GetDiagram().GetPanel().currentZoom

        # TextShape.clsLogger.debug(f'UpdateModel - ratio: {ratio}')
        if self.font is not None:
            fontSize = self.font.GetPointSize() // ratio
            self.GetModel().SetFontSize(fontSize)

    # noinspection PyUnusedLocal
    def OnLeftDown(self, event: MouseEvent):
        """
        Callback for left clicks.
        Args:
            event:
        """
        self._selected = True

    # noinspection PyUnusedLocal
    def OnLeftUp(self, event: MouseEvent):
        """
        Callback for left clicks.

        Args:
            event:
        """
        TextShape.clsLogger.debug("Unhandled left up")

    def _computeTextSize(self, dc: DC):

        textSize:       Size = dc.GetTextExtent(self.text)
        adjustedWidth:  int  = textSize.GetWidth()  + TEXT_WIDTH_ADJUSTMENT
        adjustedHeight: int  = textSize.GetHeight() + TEXT_HEIGHT_ADJUSTMENT

        self.clsLogger.debug(f'{textSize=} {adjustedWidth=} {adjustedHeight=}')
        self.SetSize(width=adjustedWidth, height=adjustedHeight)

    def _drawText(self, dc: DC):

        dc.SetBackgroundMode(PENSTYLE_SOLID)
        dc.SetTextBackground(self._textBackgroundColor)

        x, y = self.GetPosition()
        # draw the text shape with its own font
        saveFont: Font = dc.GetFont()

        if self.font is not None:
            dc.SetFont(self.font)
        dc.DrawText(self._text, x + TEXT_X_MARGIN, y + TEXT_Y_MARGIN)

        dc.SetFont(saveFont)

    def __repr__(self):
        x, y = self.GetPosition()
        return f'TextShape[{self._text=} position: ({x},{y}])'



# type: ignore

from miniogl.RectangleShape import RectangleShape

from miniogl.VShapes import VShape


class RotatableShape(RectangleShape):
    """
    Canvas for shapes that can be rotated.
    The shape is defined for one orientation, using a list of VShapes which
    is a class field named SHAPES. Then, the method Rotate can be called to
    automatically rotate the shape.

    @author Laurent Burgbacher <lb@alawa.ch>
    """
    def __init__(self, x=0.0, y=0.0, width=0.0, height=0.0, parent=None):
        """
        Constructor.

        @param double x, y : position of the point
        @param double width, height : size of the rectangle
        @param Shape parent : parent shape
        """
        RectangleShape.__init__(self, x, y, width, height, parent)
        # this is the definition of the shape
        self._defineShape()
        self._angle = 0                                 # angle is in [0..3], by steps of 90 degrees
        self._vShapes = self._SHAPES[0]                 # currently, used list of shapes
        self._InitRotations()                           # create the other rotations if necessary
        self._scale = 1.0                               # scale of the shape
        self._sox, self._soy = self._ox, self._oy       # ox, oy with scale == 1
        self._sw, self._sh = self._width, self._height  # width and height with scale == 1

    def _defineShape(self):
        """
        This is the definition of the graphical object.
        It uses a list of basic shapes, that support rotation and scaling.
        Define your own shapes in children classes by filling the innermost
        list with `VShape` instances.
        """
        self._SHAPES = [
            [
            ]
        ]

    def GetAngle(self):
        """
        Get the actual angle, in range [0; 3].

        @return int angle
        """
        return self._angle

    def SetAngle(self, angle):
        """
        Set the actual angle, in range [0; 3].
        0 is the initial angle. Each unit is a clockwise 90-degree rotation.

        @param  angle
        """
        while self._angle != angle:
            self.Rotate(True)

    def SetScale(self, scale: float):
        """
        Set the scaling of this shape.

        @param  scale
        """
        self._scale = scale
        self._ox, self._oy = self._sox * scale, self._soy * scale
        self._width, self._height = self._sw * scale, self._sh * scale

    def GetScale(self):
        """
        Get the scaling of this shape.

        @return float
        """
        return self._scale

    def SetOrigin(self, x: int, y: int):
        """
        Set the origin of the shape, from its upper left corner.
        Args:
            x: new origin abscissa
            y: new origin ordinate
        """
        self._ox = x
        self._oy = y
        scale = self._scale
        if scale != 0:
            self._sox, self._soy = x / scale, y / scale
        else:
            self._sox, self._soy = 0, 0

    def _InitRotations(self):
        """
        Init the rotations.
        Will be done just one time, or when the initial shape is changed.
        """
        if len(self._SHAPES) == 1:
            from copy import copy
            for i in range(1, 4):
                nextRotates = []
                for shape in self._SHAPES[0]:
                    n = copy(shape)
                    n.SetAngle(i)
                    nextRotates.append(n)
                self._SHAPES.append(nextRotates)

    def Rotate(self, clockwise: bool):
        """
        Rotate the shape 90 degrees clockwise or counterclockwise.

        @param  clockwise
        """
        if clockwise:
            self._angle += 1
        else:
            self._angle -= 1
        self._angle %= 4
        self._vShapes = self._SHAPES[self._angle]
        for child in self._anchors + self._children:
            lock = False
            x, y = child.GetRelativePosition()
            if not child.IsDraggable():
                child.SetDraggable(True)
                lock = True
            x, y = VShape().Convert(1, x, y)
            child.SetRelativePosition(x, y)
            if lock:
                child.SetDraggable(False)
        self._width, self._height = VShape().Convert(1, self._width, self._height)
        self._ox, self._oy = VShape().Convert(1, self._ox, self._oy)

    def Draw(self, dc, withChildren=True):
        """
        Draw the shape on the dc.

        @param  dc
        @param withChildren
        """
        if self._visible:
            RectangleShape.Draw(self, dc, False)
            for shape in self._vShapes:
                shape.Draw(dc, self._x, self._y, self._scale)
            if withChildren:
                self.DrawChildren(dc)




from logging import Logger
from logging import getLogger
from typing import List
from typing import cast

from hasiicommon.ui.widgets.DimensionsControl import DimensionsControl
from wx import Choice
from wx import CommandEvent
from wx import EVT_CHOICE
from wx import EVT_TEXT
from wx import ID_ANY
from wx import StaticText
from wx import TextCtrl
from wx import Window

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

from miniogl.MiniOglColorEnum import MiniOglColorEnum
from ogl.OglDimensions import OglDimensions
from ogl.preferences.OglPreferences import OglPreferences


class ClassAttributesControl(SizedPanel):

    def __init__(self, parent: Window):

        self.logger:       Logger          = getLogger(__name__)
        self._preferences: OglPreferences  = OglPreferences()
        super().__init__(parent)

        self._className:            TextCtrl          = cast(TextCtrl, None)
        self._classDimensions:      DimensionsControl = cast(DimensionsControl, None)
        self._classBackgroundColor: Choice            = cast(Choice, None)
        self._classTextColor:       Choice            = cast(Choice, None)

        self.SetSizerType('vertical')
        self._layoutControls(self)
        self._setControlValues()

        parent.Bind(EVT_TEXT,   self._classNameChanged,              self._className)
        parent.Bind(EVT_CHOICE, self._onClassBackgroundColorChanged, self._classBackgroundColor)
        parent.Bind(EVT_CHOICE, self._onClassTextColorChanged,       self._classTextColor)

        self.Fit()
        self.SetMinSize(self.GetSize())

    def _layoutControls(self, parentPanel: SizedPanel):

        self._layoutNameControl(parentPanel)

        self._layoutColorControls(parentPanel)

        self._classDimensions = DimensionsControl(sizedPanel=parentPanel, displayText='Class Width/Height',
                                                  valueChangedCallback=self._onClassDimensionsChanged,
                                                  setControlsSize=False)
        # noinspection PyUnresolvedReferences
        self._classDimensions.SetSizerProps(proportion=1, expand=True)

    def _layoutColorControls(self, parentPanel: SizedPanel):

        colorPanel: SizedPanel = SizedPanel(parentPanel)
        colorPanel.SetSizerType('horizontal')
        colorPanel.SetSizerProps(proportion=1, expand=False)

        self._layoutClassBackgroundControl(parentPanel=colorPanel)
        self._layoutClassTextColorControl(parentPanel=colorPanel)

    def _layoutNameControl(self, parentPanel):

        nameSizedPanel: SizedPanel = SizedPanel(parentPanel)
        nameSizedPanel.SetSizerType('form')
        nameSizedPanel.SetSizerProps(proportion=1, expand=False)

        StaticText(nameSizedPanel, ID_ANY, 'Default Class Name:')
        self._className = TextCtrl(nameSizedPanel, value=self._preferences.className, size=(160, -1))
        self._className.SetSizerProps(proportion=1, expand=False)

    def _layoutClassBackgroundControl(self, parentPanel: SizedPanel):

        classBackgroundColors = [s.value for s in MiniOglColorEnum]

        classBackgroundSSB: SizedStaticBox = SizedStaticBox(parentPanel, label='Class Background')
        classBackgroundSSB.SetSizerProps(expand=True, proportion=1)

        self._classBackgroundColor = Choice(classBackgroundSSB, choices=classBackgroundColors)

    def _layoutClassTextColorControl(self, parentPanel: SizedPanel):

        classTextColors = [s.value for s in MiniOglColorEnum]

        classTextColorSSB: SizedStaticBox = SizedStaticBox(parentPanel, label='Class Text Color')
        classTextColorSSB.SetSizerProps(expand=True, proportion=1)

        self._classTextColor = Choice(classTextColorSSB, choices=classTextColors)

    def _setControlValues(self):
        """
        """
        self._classDimensions.dimensions = self._preferences.classDimensions

        oglColors: List[str] = self._classBackgroundColor.GetItems()
        bgColorSelIdx:  int       = oglColors.index(self._preferences.classBackgroundColor.value)
        self._classBackgroundColor.SetSelection(bgColorSelIdx)

        txtColorSelIdx: int = oglColors.index(self._preferences.classTextColor.value)
        self._classTextColor.SetSelection(txtColorSelIdx)

    def _classNameChanged(self, event: CommandEvent):
        newValue: str = event.GetString()
        self._preferences.className = newValue

    def _onClassDimensionsChanged(self, newValue: OglDimensions):
        self._preferences.classDimensions = newValue

    def _onClassBackgroundColorChanged(self, event: CommandEvent):

        colorValue:    str             = event.GetString()
        oglColorEnum: MiniOglColorEnum = MiniOglColorEnum(colorValue)

        self._preferences.classBackgroundColor = oglColorEnum

    def _onClassTextColorChanged(self, event: CommandEvent):

        colorValue:    str             = event.GetString()
        oglColorEnum: MiniOglColorEnum = MiniOglColorEnum(colorValue)

        self._preferences.classTextColor = oglColorEnum




from wx import DC

from pyutmodel.PyutLink import PyutLink

from ogl.OglAssociation import OglAssociation
from ogl.OglClass import OglClass


class OglComposition(OglAssociation):
    """
    Graphical link representation of composition, (plain diamond, arrow).
    To get a new link, you should use the `OglLinkFactory` and specify
    the kind of link you want, OGL_COMPOSITION for an instance of this class.
    """

    def __init__(self, srcShape: OglClass, pyutLink: PyutLink, dstShape: OglClass, srcPos=None, dstPos=None):
        """

        Args:
            srcShape:   Source shape
            pyutLink:   Conceptual link associated with the graphical links
            dstShape:    Destination shape
            srcPos:     Position of source      Override location of input source
            dstPos:     Position of destination Override location of input destination
        """

        super().__init__(srcShape, pyutLink, dstShape, srcPos=srcPos, dstPos=dstPos)
        self.SetDrawArrow(True)

    def Draw(self, dc: DC, withChildren: bool = False):
        """

        Args:
            dc:     Device Context
            withChildren:   Draw the children or not
        """
        super().Draw(dc, withChildren=withChildren)
        self.drawDiamond(dc, True)

    def __repr__(self):
        from ogl.OglLink import OglLink
        return f'OglComposition - {OglLink.__repr__(self)}'




from typing import cast
from typing import List
from typing import NewType
from typing import Tuple
from typing import Union

from logging import Logger
from logging import getLogger

from deprecated import deprecated
from wx import BLACK_PEN
from wx import DC
from wx import RED_PEN

from miniogl.Common import Common
from miniogl.Common import CommonLine
from miniogl.Common import CommonPoint
from miniogl.LinePoint import LinePoint
from miniogl.Shape import Shape
from miniogl.AnchorPoint import AnchorPoint
from miniogl.ControlPoint import ControlPoint

ControlPoints = NewType('ControlPoints', List[LinePoint])
SegmentPoint  = NewType('SegmentPoint', Tuple[int, int])
Segments      = NewType('Segments', List[SegmentPoint])


class LineShape(Shape, Common):
    """
    This is a line, passing through control points.
    The line must begin and end with AnchorPoints, and can be guided by
    ControlPoints.
    """
    lineShapeLogger: Logger = getLogger(__name__)

    def __init__(self, srcAnchor: AnchorPoint, dstAnchor: AnchorPoint):
        """

        Args:
            srcAnchor: the source anchor of the line.
            dstAnchor: the destination anchor of the line.
        """
        Shape.__init__(self)
        self._srcAnchor: AnchorPoint = srcAnchor
        self._dstAnchor: AnchorPoint = dstAnchor

        self._controls:  ControlPoints = ControlPoints([])
        self._drawArrow: bool = True
        self._arrowSize: int = 8
        self._spline:    bool = False
        if srcAnchor:
            srcAnchor.AddLine(self)
        if dstAnchor:
            dstAnchor.AddLine(self)

    @property
    def sourceAnchor(self) -> AnchorPoint:
        return self._srcAnchor

    @sourceAnchor.setter
    def sourceAnchor(self, theNewValue: AnchorPoint):
        self._srcAnchor = theNewValue

    @property
    def destinationAnchor(self) -> AnchorPoint:
        return self._dstAnchor

    @destinationAnchor.setter
    def destinationAnchor(self, theNewValue: AnchorPoint):
        self._dstAnchor = theNewValue

    @property
    def segments(self) -> Segments:
        """
        The source anchor is the first, The destination anchor is the last.   The
        control points if any are the intermediate SegmentPoint

        Returns:  The SegmentPoint describe the line, including the intermediate control points
        where the line bends
        """
        sp: SegmentPoint = cast(SegmentPoint, self._srcAnchor.GetPosition())
        dp: SegmentPoint = cast(SegmentPoint, self._dstAnchor.GetPosition())
        LineShape.lineShapeLogger.debug(f'{self._controls=}')
        segments: Segments = Segments([])
        segments.append(sp)
        for cp in self._controls:
            lp: LinePoint = cast(LinePoint, cp)
            segments.append(cast(SegmentPoint, lp.GetPosition()))
        segments.append(dp)
        return segments

    def SetSpline(self, state):
        """
        Use a spline instead of a line.

        @param bool state : True for a spline
        """
        self._spline = state

    def GetSpline(self):
        """
        Return True if a spline is drawn instead of a line.

        @return boolean
        """
        return self._spline

    def GetPosition(self):
        """
        Return the absolute position of the shape.
        It's in the diagram's coordinate system.
        For a line, it's the middle control point, or the middle of the two
        middle control points if there's an even number of those.

        @return (double, double)
        """
        points = self.segments
        middle = len(points) // 2
        if len(points) % 2 == 0:
            # even number of points, take the two at the center
            sx, sy = points[middle-1]
            dx, dy = points[middle]
            return (sx + dx) // 2, (sy + dy) // 2
        else:
            # odd number, take the middle point
            return points[middle]

    def AddControl(self, control: ControlPoint, after: Union[ControlPoint, LinePoint] | None):
        """
        Add a control point to the line.
        The control point can be appended (last before the destination anchor)
        or inserted after a given control point.

        Args:
            control:  control point to add

            after:  This can be a control point after which to insert. This
            can also be the source of destination anchors. If it's the
            destination anchor, the point will be inserted BEFORE it.

        """
        if after is not None:
            if after is self._srcAnchor:
                self._controls.insert(0, control)
            elif after is self._dstAnchor:
                self._controls.append(control)
            else:
                i = self._controls.index(after)
                self._controls.insert(i+1, control)
        else:
            self._controls.append(control)
        control.AddLine(self)
        # add the point to the diagram so that it can be selected
        if self._diagram is not None:
            self._diagram.AddShape(control)

    @deprecated(reason='Use .destinationAnchor property')
    def GetDestination(self):
        """
        Get the destination anchor.

        @return AnchorPoint
        """
        return self._dstAnchor

    @deprecated(reason='Use .destinationAnchor property')
    def SetDestination(self, anchor):
        """
        Set the destination anchor.
        Note that the line must be removed from the previous destination!

        @param anchor
        """
        self._dstAnchor = anchor
        anchor.AddLine(self)

    def SetDrawArrow(self, draw: bool):
        """
        Set to True if you want to have an arrow head at the destination.

        @param  draw
        """
        self._drawArrow = draw

    def GetDrawArrow(self):
        """
        Tells if an arrow head will be drawn.

        @return bool
        """
        return self._drawArrow

    def SetArrowSize(self, size):
        """
        Set the size of the arrow head, in pixels.

        @param  size
        """
        self._arrowSize = size

    def GetArrowSize(self):
        """
        Get the size of the arrow head, in pixels.

        @return double size
        """
        return self._arrowSize

    @deprecated(reason='Use .sourceAnchor property')
    def GetSource(self):
        """
        Get the source anchor.

        @return AnchorPoint source
        """
        return self._srcAnchor

    @deprecated(reason='Use .sourceAnchor property')
    def SetSource(self, anchor):
        """
        Set the source anchor.
        Note that the line must be removed from the previous source!

        @param anchor
        """
        self._srcAnchor = anchor
        anchor.AddLine(self)

    @deprecated(reason='Use the .segments property')
    def GetSegments(self):
        """
        Return a list of tuples which are the coordinates of the control points.

        Returns:  A list of int tuples

        """
        sp = self._srcAnchor.GetPosition()
        dp = self._dstAnchor.GetPosition()
        # LineShape.lineShapeLogger.debug(f'GetSegments --  sp: {sp} dp: {dp}')
        retList = [sp] + list(map(lambda x: x.GetPosition(), self._controls)) + [dp]
        return retList

    def GetControlPoints(self) -> ControlPoints:
        """
        This is a copy of the original list, modifying it won't change the
        line, but modifying the control points will !

        Returns:  a list of the control points.
        """
        return ControlPoints(self._controls[:])

    def Draw(self, dc: DC, withChildren: bool = True):
        """
        Draw the line on the dc.

        Args:
            dc:
            withChildren:
        """
        if self._visible:

            super().Draw(dc=dc, withChildren=withChildren)

            line = self.segments

            if self._selected:
                dc.SetPen(RED_PEN)
            if self._spline:
                dc.DrawSpline(line)
            else:
                dc.DrawLines(line)
            for control in self._controls:
                control.Draw(dc)
            if self._selected:
                self._srcAnchor.Draw(dc)
                self._dstAnchor.Draw(dc)
            dc.SetPen(BLACK_PEN)
            if self._drawArrow:
                u, v = line[-2], line[-1]
                self.DrawArrow(dc, u, v)
            if withChildren:
                # LineShape.clsLogger.debug(f'Call DrawChildren()')
                self.DrawChildren(dc)

    def DrawBorder(self, dc):
        """
        Draw the border of the shape, for fast rendering.
        """
        self.Draw(dc)

    def DrawArrow(self, dc: DC, u: Tuple[int, int], v: Tuple[int, int]):
        """
        Draw an arrow at the end of the segment uv.

        Args:
            dc:
            u: points of the segment
            v: points of the segment
        """
        from math import pi, atan, cos, sin

        pi_6: float = pi / 6
        x1, y1 = u
        x2, y2 = v
        a = x2 - x1
        b = y2 - y1
        if abs(a) < 0.01:   # vertical segment
            if b > 0:
                alpha: float = -pi / 2
            else:
                alpha = pi / 2
        else:
            if a == 0:
                alpha = pi / 2  # TODO ?
            else:
                alpha = atan(b/a)
        if a > 0:
            alpha += pi
        alpha1: float = alpha + pi_6
        alpha2: float = alpha - pi_6
        size:   int   = self._arrowSize

        points: List[Tuple[int, int]] = [
            (x2 + round(size * cos(alpha1)), y2 + round(size * sin(alpha1))),
            (x2, y2),
            (x2 + round(size * cos(alpha2)), y2 + round(size * sin(alpha2)))
                                         ]

        dc.DrawPolygon(points)

    def Detach(self):
        """
        Detach the line and all its line points, including src and dst.
        """
        if self._diagram is not None and not self._protected:
            Shape.Detach(self)
            # while loop, because Detach will remove controls from the list on
            # which we iterate
            while self._controls:
                self._controls[0].Detach()
            self._srcAnchor.RemoveLine(self)
            self._dstAnchor.RemoveLine(self)

    def _removeControl(self, control: ControlPoint):
        """
        Remove a control point from the line.

        Args:
            control:
        """
        if control in self._controls:
            self._controls.remove(control)

    # noinspection PyUnusedLocal
    def _RemoveAnchor(self, anchor):
        """
        Remove an anchor point.

        @param anchor
        """
        self.Detach()

    def RemoveAllControlPoints(self):
        """
        Remove all the control points of the line.
        """
        while self._controls:
            self._removeControl(self._controls[0])   # type: ignore

    def Remove(self, point):
        """
        Remove a point from the line, either an anchor or control.
        If the remove point is an anchor, the line itself will be detached.

        Args:
            point: The point
        """
        if isinstance(point, AnchorPoint):
            self._RemoveAnchor(point)
        elif isinstance(point, ControlPoint):
            self._removeControl(point)

    def Inside(self, x: int, y: int) -> bool:
        """
        A tolerance of 4 pixels is used.

        Args:
            x:  abscissa click point
            y:  ordinate click point

        Returns: True if (x, y) is inside the line.
        """
        # Go through each segment of the line
        points: ControlPoints = self._mergeControlPoints()

        while len(points) > 1:

            x1, y1 = points[0].GetPosition()
            x2, y2 = points[1].GetPosition()

            startPoint: CommonPoint = CommonPoint(x=x1, y=y1)
            endPoint:   CommonPoint = CommonPoint(x=x2, y=y2)
            checkLine:  CommonLine  = CommonLine(start=startPoint, end=endPoint)

            clickDiffStartX, clickDiffStartY, diffX, diffY = self.setupInsideCheck(clickPointX=x, clickPointY=y, line=checkLine)
            # points = points[1:]
            points.pop(0)       # do this method instead of above to quiesce mypy

            if self.insideBoundingBox(clickDiffStartX, clickDiffStartY, diffX, diffY) and self.insideSegment(clickDiffStartX, clickDiffStartY, diffX, diffY):
                return True

        return False

    def _mergeControlPoints(self) -> ControlPoints:
        """
        points: ControlPoints = [self._srcAnchor] + self._controls + [self._dstAnchor]
        do this method instead of above to quiesce mypy

        Returns:  All the shape control points
        """
        points: ControlPoints = ControlPoints([])

        points.append(self._srcAnchor)

        for cp in self._controls:
            points.append(cp)

        points.append(self._dstAnchor)

        return points

    def SetSelected(self, state: bool = True):
        """
        Select the shape (default) or not

        Args:
            state:
        """
        # Shape.selected = state
        self._selected = state
        for ctrl in self._controls:
            ctrl.SetVisible(state)



from typing import Callable
from typing import cast
from typing import List
from typing import NewType
from typing import Tuple

from logging import Logger
from logging import getLogger
from logging import DEBUG

from math import pi
from math import atan
from math import cos
from math import sin

from wx import BLACK_BRUSH
from wx import BLACK_PEN
from wx import DC
from wx import FONTFAMILY_DEFAULT
from wx import FONTSTYLE_NORMAL
from wx import FONTWEIGHT_NORMAL
from wx import WHITE_BRUSH

from wx import Font

from pyutmodel.PyutLink import PyutLink

from miniogl.LineShape import Segments
from miniogl.TextShape import TextShape

from ogl.OglAssociationLabel import OglAssociationLabel
from ogl.OglLink import OglLink
from ogl.OglPosition import OglPosition

from ogl.preferences.OglPreferences import OglPreferences

DiamondPoint    = NewType('DiamondPoint', Tuple[int, int])
DiamondPoints   = NewType('DiamondPoints', List[DiamondPoint])
"""
The compute function calculates a position for a cardinality label
The parameters are the dx,dy of the source and anchor points and the link length
returns OglPosition
"""
ComputeFunction = Callable[[int, int, int], OglPosition]

PI_6:         float = pi / 6


class OglAssociation(OglLink):

    clsDiamondSize: int = OglPreferences().associationDiamondSize
    """
    Graphical link representation of an association, (simple line, no arrow).
    To get a new link,  use the `OglLinkFactory` and specify
    the link type.  .e.g. OGL_ASSOCIATION for an instance of this class.
    """
    def __init__(self, srcShape, pyutLink, dstShape, srcPos=None, dstPos=None):
        """

        Args:
            srcShape:   Source shape
            pyutLink:   Conceptual links associated with the graphical links.
            dstShape:   Destination shape
            srcPos:     Position of source      Override location of input source
            dstPos:     Position of destination Override location of input destination
        """
        self.oglAssociationLogger: Logger         = getLogger(__name__)
        self._preferences:         OglPreferences = OglPreferences()

        super().__init__(srcShape, pyutLink, dstShape, srcPos=srcPos, dstPos=dstPos)

        self._associationNameLabel:   OglAssociationLabel = OglAssociationLabel(text=self._link.name, oglPosition=OglPosition(x=0, y=0))
        self._sourceCardinality:      OglAssociationLabel = OglAssociationLabel()
        self._destinationCardinality: OglAssociationLabel = OglAssociationLabel()

        self._defaultFont: Font = Font(self._preferences.associationTextFontSize, FONTFAMILY_DEFAULT, FONTSTYLE_NORMAL, FONTWEIGHT_NORMAL)

        self._associationNameTextShape:        TextShape = cast(TextShape, None)
        self._sourceCardinalityTextShape:      TextShape = cast(TextShape, None)
        self._destinationCardinalityTextShape: TextShape = cast(TextShape, None)

        self.SetDrawArrow(False)

    @property
    def pyutObject(self) -> PyutLink:
        """
        Override
        Returns:  The data model
        """
        return self._link

    @pyutObject.setter
    def pyutObject(self, pyutLink: PyutLink):
        """
        Override in order to update the UI
        Args:
            pyutLink:
        """
        self.oglAssociationLogger.debug(f'{pyutLink=}')
        self._link = pyutLink
        self.centerLabel.text            = pyutLink.name
        self.sourceCardinality.text      = pyutLink.sourceCardinality
        self.destinationCardinality.text = pyutLink.destinationCardinality

        self.oglAssociationLogger.debug(f'{self.centerLabel=}')

    @property
    def centerLabel(self) -> OglAssociationLabel:
        return self._associationNameLabel

    @centerLabel.setter
    def centerLabel(self, newValue: OglAssociationLabel):
        self._associationNameLabel = newValue

    @property
    def sourceCardinality(self) -> OglAssociationLabel:
        return self._sourceCardinality

    @sourceCardinality.setter
    def sourceCardinality(self, newValue: OglAssociationLabel):
        self._sourceCardinality = newValue

    @property
    def destinationCardinality(self) -> OglAssociationLabel:
        return self._destinationCardinality

    @destinationCardinality.setter
    def destinationCardinality(self, newValue: OglAssociationLabel):
        self._destinationCardinality = newValue

    def Draw(self, dc: DC, withChildren: bool = True):
        """
        Called to draw the link content.
        We are going to draw all of our stuff, cardinality, Link name, etc.

        Args:
            dc:     Device context
            withChildren:   draw the children or not
        """
        OglLink.Draw(self, dc, withChildren)
        sp: Tuple[int, int] = self._srcAnchor.GetPosition()
        dp: Tuple[int, int] = self._dstAnchor.GetPosition()

        oglSp: OglPosition = OglPosition(x=sp[0], y=sp[1])
        oglDp: OglPosition = self._computeDestinationPosition(sp=OglPosition.tupleToOglPosition(sp), dp=OglPosition.tupleToOglPosition(dp))

        if self._link.name != '' and self._associationNameTextShape is None:
            self._createAssociationName()
        elif self._associationNameTextShape is not None:
            self._updateAssociationState()

        if self._link.sourceCardinality != '' and self._sourceCardinalityTextShape is None:
            self._createSourceCardinality(sp=oglSp)
        elif self._sourceCardinalityTextShape is not None:
            self._updateSourceCardinalityState()

        if self._link.destinationCardinality != '' and self._destinationCardinalityTextShape is None:
            self._createDestinationCardinality(dp=oglDp)
        elif self._destinationCardinalityTextShape is not None:
            self._updateDestinationCardinalityState()

    def drawDiamond(self, dc: DC, filled: bool = False):
        """
        Draw an arrow at the beginning of the line.

        Args:
            dc:         The device context
            filled:     True if the diamond must be filled, False otherwise
        """
        #
        line: Segments = self.segments

        # self.oglAssociationLogger.debug(f'{line=}')
        points: DiamondPoints = OglAssociation.calculateDiamondPoints(lineSegments=line)
        # self.oglAssociationLogger.debug(f'{points:}')

        dc.SetPen(BLACK_PEN)
        if filled:
            dc.SetBrush(BLACK_BRUSH)
        else:
            dc.SetBrush(WHITE_BRUSH)
        dc.DrawPolygon(points)
        dc.SetBrush(WHITE_BRUSH)

    def _createAssociationName(self):
        """
        Create association name text shape;
        """
        self._associationNameTextShape = self._createTextShapeFromAssociationLabel(associationLabel=self._associationNameLabel)

    def _createSourceCardinality(self, sp: OglPosition):

        self._sourceCardinalityTextShape = self._createTextShapeFromAssociationLabel(associationLabel=self._sourceCardinality)

        srcX, srcY = self._sourceCardinalityTextShape.ConvertCoordToRelative(x=sp.x, y=sp.y)
        self._sourceCardinalityTextShape.SetRelativePosition(x=srcX, y=srcY)
        self._sourceCardinality.oglPosition = OglPosition(x=srcX, y=srcY)

    def _createDestinationCardinality(self, dp: OglPosition):
        self._destinationCardinalityTextShape = self._createTextShapeFromAssociationLabel(associationLabel=self._destinationCardinality)

        dstX, dstY = self._destinationCardinalityTextShape.ConvertCoordToRelative(x=dp.x, y=dp.y)
        self._destinationCardinalityTextShape.SetRelativePosition(x=dstX, y=dstY)
        self._destinationCardinality.oglPosition = OglPosition(x=dstX, y=dstY)

    def _createTextShapeFromAssociationLabel(self, associationLabel: OglAssociationLabel) -> TextShape:

        oglPosition:    OglPosition = associationLabel.oglPosition
        labelTextShape: TextShape   = self._createTextShape(x=oglPosition.x, y=oglPosition.y, text=associationLabel.text, font=self._defaultFont)

        labelTextShape.draggable = True

        return labelTextShape

    def _updateAssociationState(self):

        x, y = self._associationNameTextShape.GetRelativePosition()

        self._associationNameLabel.oglPosition = OglPosition(x=x, y=y)
        self._associationNameTextShape.text = self._link.name

    def _updateSourceCardinalityState(self):
        x, y = self._sourceCardinalityTextShape.GetRelativePosition()

        self._sourceCardinality.oglPosition = OglPosition(x=x, y=y)
        self._sourceCardinalityTextShape.text = self._link.sourceCardinality

    def _updateDestinationCardinalityState(self):
        x, y = self._sourceCardinalityTextShape.GetRelativePosition()

        self.destinationCardinality.oglPosition = OglPosition(x=x, y=y)
        self._destinationCardinalityTextShape.text = self._link.destinationCardinality

    def _computeDestinationPosition(self, sp: OglPosition, dp: OglPosition) -> OglPosition:

        def computeDestination(dx, dy, linkLength) -> OglPosition:
            x: int = round((-20 * dx / linkLength + dx * 5 / linkLength) + dp.x)
            y: int = round((-20 * dy / linkLength - dy * 5 / linkLength) + dp.y)

            return OglPosition(x=x, y=y)

        return self._computePosition(sp=sp, dp=dp, computeFunction=computeDestination)

    def _computeSourcePosition(self, sp: OglPosition, dp: OglPosition) -> OglPosition:

        def computeSource(dx, dy, linkLength):
            x: int = round((20 * dx / linkLength - dx * 5 / linkLength) + sp.x)
            y: int = round((20 * dy / linkLength + dy * 5 / linkLength) + sp.y)

            return OglPosition(x=x, y=y)

        return self._computePosition(sp=sp, dp=dp, computeFunction=computeSource)

    def _computePosition(self, sp: OglPosition, dp: OglPosition, computeFunction: ComputeFunction) -> OglPosition:

        dx, dy          = self._computeDxDy(srcPosition=sp, destPosition=dp)
        linkLength: int = self._computeLinkLength(srcPosition=sp, destPosition=dp)

        oglPosition: OglPosition = computeFunction(dx, dy, linkLength)

        if self.oglAssociationLogger.isEnabledFor(DEBUG):
            info = (
                f'{sp=} '
                f'{dp=} '
                f'{dx=} '
                f'{dy=} '
                f'linkLength={linkLength:.2f} '
                f'{oglPosition=}'
            )
            self.oglAssociationLogger.debug(info)
        return oglPosition

    @staticmethod
    def calculateDiamondPoints(lineSegments: Segments) -> DiamondPoints:
        """
        Made static so that we can unit test it;  Please the only instance variables needed
        are passed in

        Args:
            lineSegments:  The line where we are putting the diamondPoints

        Returns:  The diamond points that define the diamond polygon
        """
        x1, y1 = lineSegments[1]
        x2, y2 = lineSegments[0]
        a: int = x2 - x1
        b: int = y2 - y1
        if abs(a) < 0.01:  # vertical segment
            if b > 0:
                alpha: float = -pi / 2
            else:
                alpha = pi / 2
        else:
            if a == 0:
                if b > 0:
                    alpha = pi / 2
                else:
                    alpha = 3 * pi / 2
            else:
                alpha = atan(b/a)
        if a > 0:
            alpha += pi
        alpha1: float = alpha + PI_6
        alpha2: float = alpha - PI_6

        diamondPoints: DiamondPoints = DiamondPoints([])

        dp0: DiamondPoint = OglAssociation.calculateDiamondPoint0(x2=x2, y2=y2, alpha1=alpha1)
        diamondPoints.append(dp0)

        diamondPoints.append(DiamondPoint((x2, y2)))

        dp2: DiamondPoint = OglAssociation.calculateDiamondPoint2(x2=x2, y2=y2, alpha2=alpha2)
        diamondPoints.append(dp2)

        dp3: DiamondPoint = OglAssociation.calculateDiamondPoint3(x2=x2, y2=y2, alpha=alpha)
        diamondPoints.append(dp3)

        return diamondPoints

    @classmethod
    def calculateDiamondPoint0(cls, x2: float, y2: float, alpha1: float) -> DiamondPoint:

        dpx0: float = x2 + OglAssociation.clsDiamondSize * cos(alpha1)
        dpy0: float = y2 + OglAssociation.clsDiamondSize * sin(alpha1)

        return DiamondPoint((round(dpx0), round(dpy0)))

    @classmethod
    def calculateDiamondPoint2(cls, x2: float, y2: float, alpha2: float) -> DiamondPoint:

        dpx2: float = x2 + OglAssociation.clsDiamondSize * cos(alpha2)
        dpy2: float = y2 + OglAssociation.clsDiamondSize * sin(alpha2)

        return DiamondPoint((round(dpx2), round(dpy2)))

    @classmethod
    def calculateDiamondPoint3(cls, x2: float, y2: float, alpha: float) -> DiamondPoint:

        dpx3: float = x2 + 2 * OglAssociation.clsDiamondSize * cos(alpha)
        dpy3: float = y2 + 2 * OglAssociation.clsDiamondSize * sin(alpha)

        return DiamondPoint((round(dpx3), round(dpy3)))

    def __repr__(self):
        return f'OglAssociation - {super().__repr__()}'




from logging import Logger
from logging import getLogger

from wx import DC

from ogl.OglAssociation import OglAssociation


class OglAggregation(OglAssociation):
    """
    Graphical link representation of aggregation, (empty diamond, arrow).
    To get a new link, you should use the `OglLinkFactory` and specify
    the kind of link you want, OGL_AGGREGATION for an instance of this class.
    """

    def __init__(self, srcShape, pyutLink, dstShape, srcPos=None, dstPos=None):
        """

        Args:
            srcShape:   Source shape
            pyutLink:   Conceptual links associated with the graphical links.
            dstShape:   Destination shape
            srcPos:     Position of source      Override location of input source
            dstPos:     Position of destination Override location of input destination
        """
        super().__init__(srcShape, pyutLink, dstShape, dstPos=dstPos, srcPos=srcPos)

        self.logger: Logger = getLogger(__name__)
        self.SetDrawArrow(True)

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Called to draw link contents
        Args:
            dc:     The device context
            withChildren:  Should we draw children
        """
        super().Draw(dc, withChildren)

        # Draw diamond
        self.drawDiamond(dc, False)

    def __repr__(self):
        from ogl.OglLink import OglLink
        return f'OglAggregation - {OglLink.__repr__(self)}'




from typing import List
from typing import cast

from logging import Logger
from logging import getLogger

from miniogl.MiniOglColorEnum import MiniOglColorEnum
from miniogl.MiniOglPenStyle import MiniOglPenStyle

from wx import CB_READONLY
from wx import EVT_CHECKBOX
from wx import EVT_CHOICE
from wx import EVT_COMBOBOX
from wx import EVT_SPINCTRL

from wx import CheckBox
from wx import Choice
from wx import ComboBox
from wx import SpinCtrl
from wx import SpinEvent
from wx import CommandEvent
from wx import Window

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

from ogl.ui.BaseOglPreferencesPage import BaseOglPreferencesPage


SPINNER_WIDTH:  int = 60
SPINNER_HEIGHT: int = 35


class DiagramPreferencesPage(BaseOglPreferencesPage):
    """
    This is a complex layout.  The following description is good as of version 0.60.50;
    If you change the layout please update this (Yeah, I do not like this kind of brittle
    documentation either

        ----------------- Dialog Sized Panel --------------------------------------------------------------------------------------------------------
        |                                                                                                                                           |
        |   ----------------- HorizontalPanel ---------------------------------------------------------------------------------------------         |
        |  |     ------------------- Vertical Panel --------------   ------------------- gridIntervalSSB -------------------              |         |
        |  |     |                                               |   |                                                      |             |         |
        |  |     |                                               |   |                                                      |             |         |
        |  |     |                                               |   |                                                      |             |         |
        |  |     |                                               |   |                                                      |             |         |
        |  |     |                                               |   |                                                      |             |         |
        |  |     -------------------------------------------------   --------------------------------------------------------             |         |
        |  --------------------------------------------------------------------------------------------------------------------------------         |
        |                                                                                                                                           |
        |   ------------------------------------------gridPanel-------------------------------------------------------------------                  |
        |  |                                                                                                                     |                  |
        |  |    ---------- gridLineColorSSB ------------------     ---------- gridLineStyleSSB ------------------                |                  |
        |  |    |                                             |   |                                             |                |                  |
        |  |    |                                             |   |                                             |                |                  |
        |  |    |                                             |   |                                             |                |                  |
        |  |    |                                             |   |                                             |                |                  |
        |  |    |                                             |   |                                             |                |                  |
        |  |    |                                             |   |                                             |                |                  |
        |  |    ---------- gridLineColorSSB ------------------     ---------- gridLineStyleSSB ------------------                |                  |
        |  |                                                                                                                     |                  |
        |  |                                                                                                                     |                  |
        |   ----------------------------------------------------------------------------------------------------------------------                  |
        |                                                                                                                                           |
        |                                                                                                                                           |
        ----------------- Dialog Sized Panel --------------------------------------------------------------------------------------------------------

    """

    def __init__(self, parent: Window):

        self.logger: Logger = getLogger(__name__)

        super().__init__(parent)
        self.SetSizerType('vertical')

        self._enableBackgroundGrid: CheckBox = cast(CheckBox, None)
        self._snapToGrid:           CheckBox = cast(CheckBox, None)
        self._centerDiagramView:    CheckBox = cast(CheckBox, None)
        self._showParameters:       CheckBox = cast(CheckBox, None)
        self._gridInterval:         SpinCtrl = cast(SpinCtrl, None)
        self._gridLineColor:        ComboBox = cast(ComboBox, None)
        self._gridStyleChoice:      Choice   = cast(Choice, None)

        self._layoutControls(self)
        self._setControlValues()
        self._bindCallbacks(parent=self)

    def _layoutControls(self, parentSizedPanel: SizedPanel):

        horizontalPanel: SizedPanel = SizedPanel(parentSizedPanel)
        verticalPanel:   SizedPanel = SizedPanel(horizontalPanel)
        horizontalPanel.SetSizerType('horizontal')
        horizontalPanel.SetSizerProps(expand=True, proportion=3)
        verticalPanel.SetSizerType('vertical')

        self._layoutDiagramPreferences(verticalPanel=verticalPanel)

        self._layoutGridIntervalControl(sizedPanel=horizontalPanel)

        self._layoutGridOptions(panel=parentSizedPanel)

        self._fixPanelSize(panel=self)

    @property
    def name(self) -> str:
        return 'Diagram'

    def _setControlValues(self):
        """
        """
        self._resetSnapToGridControl()

        self._enableBackgroundGrid.SetValue(self._preferences.backgroundGridEnabled)
        self._snapToGrid.SetValue(self._preferences.snapToGrid)
        self._centerDiagramView.SetValue(self._preferences.centerDiagram)
        self._showParameters.SetValue(self._preferences.showParameters)

        self._gridInterval.SetValue(self._preferences.backgroundGridInterval)
        self._gridLineColor.SetValue(self._preferences.gridLineColor.value)

        gridLineStyles: List[str] = self._gridStyleChoice.GetItems()
        selectedIndex:  int       = gridLineStyles.index(self._preferences.gridLineStyle.value)
        self._gridStyleChoice.SetSelection(selectedIndex)

    def _bindCallbacks(self, parent):

        parent.Bind(EVT_CHECKBOX, self._onEnableBackgroundGridChanged,   self._enableBackgroundGrid)
        parent.Bind(EVT_CHECKBOX, self._onSnapToGridChanged,             self._snapToGrid)
        parent.Bind(EVT_CHECKBOX, self._onCenterDiagramViewChanged,      self._centerDiagramView)
        parent.Bind(EVT_CHECKBOX, self._onShowParametersChanged,         self._showParameters)
        parent.Bind(EVT_COMBOBOX, self._onGridLineColorSelectionChanged, self._gridLineColor)
        parent.Bind(EVT_SPINCTRL, self._onGridIntervalChanged,           self._gridInterval)
        parent.Bind(EVT_CHOICE,   self._onGridStyleChanged,              self._gridStyleChoice)

    def _layoutDiagramPreferences(self, verticalPanel: SizedPanel):

        self._enableBackgroundGrid = CheckBox(verticalPanel, label='Enable Background Grid')
        self._snapToGrid           = CheckBox(verticalPanel, label='Snap Shapes to Grid')
        self._centerDiagramView    = CheckBox(verticalPanel, label='Center Diagram View')
        self._showParameters       = CheckBox(verticalPanel, label='Show Method Parameters')

        self._enableBackgroundGrid.SetToolTip('Turn on a diagram grid in the UML Frame')
        self._snapToGrid.SetToolTip('Snap class diagram shapes to the closest grid corner')
        self._centerDiagramView.SetToolTip('Center the view in the virtual frame')
        self._showParameters.SetToolTip('Global value to display method parameters;  Unless overridden by the class')

    def _layoutGridIntervalControl(self, sizedPanel):
        gridIntervalSSB: SizedStaticBox = SizedStaticBox(sizedPanel, label='Grid Interval')
        gridIntervalSSB.SetSizerProps(expand=True)
        self._gridInterval = SpinCtrl(parent=gridIntervalSSB)

    def _layoutGridOptions(self, panel: SizedPanel):

        gridPanel: SizedPanel = SizedPanel(panel)
        gridPanel.SetSizerType('horizontal')
        gridPanel.SetSizerProps(expand=True, proportion=2)

        self._createGridLineColorControl(panel=gridPanel)
        self._createGridStyleChoice(panel=gridPanel)

    def _createGridLineColorControl(self, panel: SizedPanel):

        colorChoices = []
        for cc in MiniOglColorEnum:
            colorChoices.append(cc.value)

        gridLineColorSSB: SizedStaticBox = SizedStaticBox(panel, label='Grid Line Color')
        gridLineColorSSB.SetSizerProps(expand=True, proportion=1)

        self._gridLineColor = ComboBox(gridLineColorSSB, choices=colorChoices, style=CB_READONLY)

    def _createGridStyleChoice(self, panel: SizedPanel):

        gridStyles = [s.value for s in MiniOglPenStyle]

        gridLineStyleSSB: SizedStaticBox = SizedStaticBox(panel, label='Grid Line Style')
        gridLineStyleSSB.SetSizerProps(expand=True, proportion=1)

        self._gridStyleChoice = Choice(gridLineStyleSSB, choices=gridStyles)

    def _onEnableBackgroundGridChanged(self, event: CommandEvent):

        newValue: bool = event.IsChecked()
        self.logger.debug(f'onEnableBackgroundGridChanged - {newValue=}')
        self._preferences.backgroundGridEnabled = newValue
        self._resetSnapToGridControl()

    def _onSnapToGridChanged(self, event: CommandEvent):

        newValue: bool = event.IsChecked()
        self.logger.debug(f'onSnapToGridChanged - {newValue=}')
        self._preferences.snapToGrid = newValue

    def _onCenterDiagramViewChanged(self, event: CommandEvent):
        newValue: bool = event.IsChecked()
        self._preferences.centerDiagram = newValue

    def _onShowParametersChanged(self, event: CommandEvent):
        newValue: bool = event.IsChecked()
        self._preferences.showParameters = newValue

    def _onGridLineColorSelectionChanged(self, event: CommandEvent):

        colorValue:    str              = event.GetString()
        pyutColorEnum: MiniOglColorEnum = MiniOglColorEnum(colorValue)

        self._preferences.gridLineColor = pyutColorEnum

    def _onGridIntervalChanged(self, event: SpinEvent):

        newInterval: int = event.GetInt()
        self._preferences.backgroundGridInterval = newInterval

    def _onGridStyleChanged(self, event: CommandEvent):

        styleText: str = event.GetString()
        self.logger.warning(f'{styleText=}')

        pyutPenStyle: MiniOglPenStyle = MiniOglPenStyle(styleText)

        self._preferences.gridLineStyle = pyutPenStyle

    def _resetSnapToGridControl(self):
        """
        Make the UI consistent when the background grid is used or not
        If no background grid there is nothing to snap to
        """
        if self._preferences.backgroundGridEnabled is True:
            self._snapToGrid.Enabled = True
        else:
            self._snapToGrid.SetValue(False)
            self._snapToGrid.Enabled = False
            self._preferences.snapToGrid = False




from typing import Tuple

from dataclasses import dataclass

from math import sqrt


@dataclass
class CommonPoint:

    x: int = 0
    y: int = 0


@dataclass
class CommonLine:

    start: CommonPoint = CommonPoint(0, 0)
    end:   CommonPoint = CommonPoint(0, 0)


class Common:

    CLICK_TOLERANCE: float = 4.0
    """
    A `mix-in` class to share common code currently between a LineShape and a shape that is
    almost a line.  :-)

    """

    def setupInsideCheck(self, clickPointX: int, clickPointY: int, line: CommonLine) -> Tuple[int, int, int, int]:

        x1: int = line.start.x
        y1: int = line.start.y
        x2: int = line.end.x
        y2: int = line.end.y

        diffX: int = x2 - x1
        diffY: int = y2 - y1

        clickDiffStartX: int = clickPointX - x1     # x - x1
        clickDiffStartY: int = clickPointY - y1     # y - y1

        return clickDiffStartX, clickDiffStartY, diffX, diffY

    def insideSegment(self, clickDiffStartX: float, clickDiffStartY: float, diffX: float, diffY: float) -> bool:
        """

        Args:
            clickDiffStartX:
            clickDiffStartY:
            diffX:  Difference between the line's end X coordinate and the start X coordinate
            diffY:  Difference between the line's end Y coordinate and the start Y coordinate

        Returns: `True` if the click points is withing the click tolerance; else `False`
        """
        den: float = sqrt(diffX * diffX + diffY * diffY)

        if den != 0.0:
            d: float = (clickDiffStartX * diffY - clickDiffStartY * diffX) / den
        else:
            return False
        return abs(d) < Common.CLICK_TOLERANCE

    def insideBoundingBox(self, clickDiffStartX: float, clickDiffStartY: float, diffX: float, diffY: float):
        """
        Check if the point (x, y) is inside a box of origin (0, 0) and
        diagonal (a, b) with a tolerance of CLICK_TOLERANCE

        Args:
            clickDiffStartX:
            clickDiffStartY:
            diffX:  Difference between the line's end X coordinate and the start X coordinate
            diffY:  Difference between the line's end Y coordinate and the start Y coordinate

        Returns:  `True` if the click point is in the bounding box else `False`

        """
        ma: float = diffX / 2
        mb: float = diffY / 2

        if diffX > 0:
            w: int = max(4, int(diffX - 8))
        else:
            w = min(-4, int(diffX + 8))

        if diffY > 0:
            h: int = max(4, int(diffY - 8))
        else:
            h = min(-4, int(diffY + 8))

        topLeftX: float = ma - w / 2
        topLeftY: float = mb - h / 2

        i:  bool = clickDiffStartX > topLeftX
        j:  bool = clickDiffStartX > topLeftX + w
        k:  bool = clickDiffStartY > topLeftY
        ll: bool = clickDiffStartY > topLeftY + h

        ans: bool = (i + j) == 1 and (k + ll) == 1
        return ans




from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from wx import ALL
from wx import BoxSizer
from wx import Button
from wx import CANCEL
from wx import CENTER
from wx import EVT_BUTTON
from wx import EVT_CLOSE
from wx import EVT_TIMER
from wx import EXPAND
from wx import GetMouseState
from wx import MouseState
from wx import StaticText
from wx import Timer
from wx import TimerEvent
from wx import VERTICAL
from wx import HORIZONTAL
from wx import ID_ANY
from wx import ID_OK
from wx import OK

from wx import StaticBox
from wx import StaticBoxSizer
from wx import Window
from wx import Dialog
from wx import CommandEvent

from wx import NewIdRef as wxNewIdRef


class DlgDebugDiagramFrame(Dialog):
    """
    Use as follows:

    debugDialog: DlgDebugDiagramFrame = DlgDebugDiagramFrame(frameToMonitor, ID_ANY)
    debugDialog.startMonitor()
    debugDialog.Show(True)

    The self parameter is the UML frame to monitor
    """

    CONTAINER_GAP:         int = 3
    VERTICAL_GAP:          int = 2
    HORIZONTAL_GAP:        int = 3
    PROPORTION_CHANGEABLE: int = 1

    DEBUG_TIMER_UPDATE_MSECS: int = 500

    """
    Sample use:
            self._debugDialog: DlgDebugDiagramFrame = DlgDebugDiagramFrame(parent, ID_ANY)
            self._debugDialog.Show(True)
            
    """
    def __init__(self, frameToMonitor: Window, dialogIdentifier):
        """

        Args:
            frameToMonitor:     parent window to center on and the one to monitor
            dialogIdentifier:   An identifier for the dialog
        """
        super().__init__(frameToMonitor, dialogIdentifier, "Debug Diagram")

        from miniogl.DiagramFrame import DiagramFrame

        self.logger:        Logger       = getLogger(__name__)
        assert isinstance(frameToMonitor, DiagramFrame)
        self._diagramFrame: DiagramFrame = cast(DiagramFrame, frameToMonitor)

        hs:        BoxSizer = self.__createDialogButtonsContainer()
        box:       StaticBox = StaticBox(self, ID_ANY, "")

        mainSizer: StaticBoxSizer = StaticBoxSizer(box, VERTICAL)
        self.__initializeTheControls(mainSizer)
        mainSizer.Add(hs, 0, CENTER)

        border: BoxSizer = BoxSizer()
        border.Add(mainSizer, DlgDebugDiagramFrame.PROPORTION_CHANGEABLE, flag=EXPAND | ALL, border=0)

        self.SetAutoLayout(True)
        self.SetSizer(border)

        border.Fit(self)
        border.SetSizeHints(self)

        self._timer: Timer = Timer(self)

        self.Bind(EVT_BUTTON, self.__OnCmdOk, id=ID_OK)
        self.Bind(EVT_CLOSE,  self.__OnClose)
        self.Bind(EVT_TIMER,  self._onTimer, self._timer)

    def startMonitor(self):
        self._timer.Start(DlgDebugDiagramFrame.DEBUG_TIMER_UPDATE_MSECS)

    # noinspection PyUnusedLocal
    def _onTimer(self, event: TimerEvent):

        ms: MouseState = GetMouseState()

        msX: int = ms.GetX()
        msY: int = ms.GetY()
        wx, wy = self._diagramFrame.ScreenToClient(msX, msY)

        self.__x.SetLabel(str(wx))
        self.__y.SetLabel(str(wy))

    def __initializeTheControls(self, mainSizer: StaticBoxSizer):
        """
        Initialize the controls.
        """
        # IDs
        self.__xId: wxNewIdRef = wxNewIdRef()
        self.__yId: wxNewIdRef = wxNewIdRef()

        xBox, self.__x = self.__createPositionContainer('Frame X Position: ', self.__xId)
        yBox, self.__y = self.__createPositionContainer('Frame Y Position: ', self.__yId)

        mainSizer.Add(xBox, 0, ALL, DlgDebugDiagramFrame.VERTICAL_GAP)
        mainSizer.Add(yBox, 0, ALL, DlgDebugDiagramFrame.VERTICAL_GAP)

    def __createPositionContainer(self, labelText: str, posId: int) -> Tuple[BoxSizer, StaticText]:

        box: BoxSizer = BoxSizer(HORIZONTAL)
        lbl:      StaticText = StaticText(self, ID_ANY, labelText)
        reporter: StaticText = StaticText(self, posId, "00000")

        box.Add(lbl,   0, ALL, DlgDebugDiagramFrame.HORIZONTAL_GAP)
        box.Add(reporter, 0, ALL, DlgDebugDiagramFrame.HORIZONTAL_GAP)

        return box, reporter

    def __createDialogButtonsContainer(self) -> BoxSizer:

        hs: BoxSizer = BoxSizer(HORIZONTAL)

        btnOk: Button = Button(self, ID_OK, '&OK')
        hs.Add(btnOk, 0, ALL, DlgDebugDiagramFrame.CONTAINER_GAP)
        return hs

    def __OnCmdOk(self, event: CommandEvent):
        """
        """
        self._timer.Stop()
        event.Skip(skip=True)
        self.SetReturnCode(OK)
        self.EndModal(OK)

    # noinspection PyUnusedLocal
    def __OnClose(self, event: CommandEvent):
        """
        """
        self._timer.Stop()
        self.SetReturnCode(CANCEL)
        self.EndModal(CANCEL)



from typing import List

from wx import BK_DEFAULT
from wx import Bitmap
from wx import ID_ANY
from wx import ImageList
from wx import Toolbook
from wx import Window

from wx.lib.embeddedimage import PyEmbeddedImage
from wx.lib.sized_controls import SizedPanel

from hasiicommon.resources.images.DefaultPreferences import embeddedImage as DefaultPreferences

from hasiicommon.resources.images.icons.embedded16.ImgToolboxNote import embeddedImage as ImgToolboxNote
from hasiicommon.resources.images.icons.embedded16.ImgToolboxText import embeddedImage as ImgToolboxText
from hasiicommon.resources.images.icons.embedded16.ImgToolboxClass import embeddedImage as ImgToolboxClass
from hasiicommon.resources.images.icons.embedded16.ImgToolboxSequenceDiagramInstance import embeddedImage as ImgToolboxSequenceDiagramInstance
from hasiicommon.resources.images.icons.embedded16.ImgToolboxRelationshipComposition import embeddedImage as ImgToolboxRelationshipComposition

from ogl.ui.valuecontrols.AssociationAttributesControl import AssociationAttributesControl
from ogl.ui.valuecontrols.ClassAttributesControl import ClassAttributesControl
from ogl.ui.valuecontrols.DefaultNamesControl import DefaultNamesControl
from ogl.ui.valuecontrols.NoteAttributesControl import NoteAttributesControl
from ogl.ui.valuecontrols.SDAttributesControl import SDAttributesControl
from ogl.ui.valuecontrols.TextAttributesControl import TextAttributesControl

from ogl.ui.BaseOglPreferencesPage import BaseOglPreferencesPage


def getNextImageID(count):
    imID = 0
    while True:
        yield imID
        imID += 1
        if imID == count:
            imID = 0


class DefaultValuesPreferencesPage(BaseOglPreferencesPage):

    def __init__(self, parent: Window):
        super().__init__(parent)
        self._layoutWindow(self)
        self._fixPanelSize(self)

    @property
    def name(self) -> str:
        return 'Default Values'

    def _layoutWindow(self, parent: SizedPanel):

        toolBook: Toolbook = Toolbook(parent, ID_ANY, style=BK_DEFAULT)
        toolBook.SetSizerProps(expand=True, proportion=1)

        embeddedImages: List[PyEmbeddedImage] = [ImgToolboxNote, ImgToolboxText, ImgToolboxClass, DefaultPreferences,
                                                 ImgToolboxSequenceDiagramInstance,
                                                 ImgToolboxRelationshipComposition
                                                 ]
        imageList:      ImageList             = ImageList(width=16, height=16)

        for embeddedImage in embeddedImages:
            bmp: Bitmap = embeddedImage.GetBitmap()
            imageList.Add(bmp)

        toolBook.AssignImageList(imageList)

        imageIdGenerator = getNextImageID(imageList.GetImageCount())

        notePanel:  NoteAttributesControl  = NoteAttributesControl(parent=toolBook)
        textPanel:  TextAttributesControl  = TextAttributesControl(parent=toolBook)
        classPanel: ClassAttributesControl = ClassAttributesControl(parent=toolBook)

        defaultNamesPanel: DefaultNamesControl = DefaultNamesControl(parent=toolBook)
        sdPanel:           SDAttributesControl = SDAttributesControl(parent=toolBook)

        associationPanel:  AssociationAttributesControl = AssociationAttributesControl(parent=toolBook)

        toolBook.AddPage(notePanel,         text='Notes', select=True, imageId=next(imageIdGenerator))
        toolBook.AddPage(textPanel,         text='Text',  select=False, imageId=next(imageIdGenerator))
        toolBook.AddPage(classPanel,        text='Class', select=False, imageId=next(imageIdGenerator))
        toolBook.AddPage(defaultNamesPanel, text='Names', select=False, imageId=next(imageIdGenerator))
        toolBook.AddPage(sdPanel,           text='SD',    select=False, imageId=next(imageIdGenerator))
        toolBook.AddPage(associationPanel,  text='Association', select=False, imageId=next(imageIdGenerator))

    def _setControlValues(self):
        pass




from dataclasses import dataclass

from ogl.OglPosition import OglPosition


@dataclass
class OglAssociationLabel:

    text:        str         = ''
    oglPosition: OglPosition = OglPosition(x=0, y=0)




from logging import Logger
from logging import getLogger

from wx import BLACK_PEN
from wx import RED_PEN

from wx import Pen
from wx import MouseEvent

from miniogl.Shape import Shape
from miniogl.AnchorPoint import AnchorPoint
from miniogl.ShapeEventHandler import ShapeEventHandler

from miniogl.AttachmentSide import AttachmentSide


class SelectAnchorPoint(AnchorPoint, ShapeEventHandler):

    """
    This is a point attached to a shape to indicate where to click;  Presumably, to indicate where
    to attach something

    """
    def __init__(self, x: int, y: int, attachmentSide: AttachmentSide, parent: Shape | None):
        """

        Args:
            x: x position of the point
            y: y position of the point
            parent:
        """
        super().__init__(x, y, parent)

        self.logger: Logger = getLogger(__name__)
        self._attachmentSide: AttachmentSide = attachmentSide
        self._pen:             Pen                = RED_PEN
        self.SetStayInside(True)
        self.draggable = False     # So it sticks on OglClass resize;  But now the user can move it !!
        self.SetStayOnBorder(True)

    @property
    def attachmentPoint(self) -> AttachmentSide:
        return self._attachmentSide

    @attachmentPoint.setter
    def attachmentPoint(self, newValue: AttachmentSide):
        self._attachmentSide = newValue

    def setYouAreTheSelectedAnchor(self):
        self._pen = BLACK_PEN

    def Draw(self, dc, withChildren=True):

        dc.SetPen(self._pen)
        super().Draw(dc, withChildren)

    def OnLeftDown(self, event: MouseEvent):
        """
        Callback for left clicks.

        Args:
            event: The mouse event
        """
        self.logger.debug(f'SelectAnchorPoint.OnLeftDown:  {self._parent=} {event.GetPosition()=}')
        self._parent.handleSelectAnchorPointSelection(event)      # bad form;  but anything to get rid of mediator

    def __str__(self) -> str:
        x, y = self.GetPosition()
        draggable: bool = self._draggable
        return f'SelectAnchorPoint[({x},{y}) - {draggable=}]'

    def __repr__(self):
        return self.__str__()



from wx import Window

from wx.lib.sized_controls import SizedPanel

from ogl.preferences.OglPreferences import OglPreferences


class BaseOglPreferencesPage(SizedPanel):

    def __init__(self, parent: Window):

        super().__init__(parent)

        self._preferences: OglPreferences = OglPreferences()

    def _fixPanelSize(self, panel: SizedPanel):
        """
        Do the following or does not get resized correctly
        A little trick to make sure that the sizer cannot be resized to
        less screen space than the controls need

        Args:
            panel:
        """
        panel.Fit()
        panel.SetMinSize(panel.GetSize())




from logging import Logger
from logging import getLogger

from wx import EVT_TEXT
from wx import ID_ANY

from wx import CommandEvent
from wx import StaticText
from wx import TE_MULTILINE
from wx import TextCtrl
from wx import Window

from wx.lib.sized_controls import SizedPanel

from hasiihelper.Dimensions import Dimensions

from hasiicommon.ui.widgets.DimensionsControl import DimensionsControl

from ogl.OglDimensions import OglDimensions
from ogl.preferences.OglPreferences import OglPreferences


class NoteAttributesControl(SizedPanel):

    def __init__(self, parent: Window):

        self.logger:       Logger         = getLogger(__name__)
        self._preferences: OglPreferences = OglPreferences()
        super().__init__(parent)

        nameSizer: SizedPanel = SizedPanel(self)
        nameSizer.SetSizerProps(proportion=0, expand=False)

        StaticText(nameSizer, ID_ANY, 'Default Note Text:')
        noteText: TextCtrl = TextCtrl(nameSizer, value=self._preferences.noteText, size=(400, 100), style=TE_MULTILINE)
        noteText.SetSizerProps(expand=True, proportion=1)
        parent.Bind(EVT_TEXT, self._onNoteTextChanged, noteText)

        self._noteDimensions: DimensionsControl = DimensionsControl(sizedPanel=self, displayText='Note Width/Height',
                                                                    valueChangedCallback=self._noteDimensionsChanged,
                                                                    setControlsSize=False)

        self._noteDimensions.dimensions = self._preferences.noteDimensions

    def _onNoteTextChanged(self, event: CommandEvent):
        newText: str = event.GetString()
        self._preferences.noteText = newText

    def _noteDimensionsChanged(self, newValue: Dimensions):

        # Just as easy to just cast this;  But, I want no technical debt
        self._preferences.noteDimensions = OglDimensions(width=newValue.width, height=newValue.height)



# ----------------------------------------------------------------------
# This file was generated by /Users/humberto.a.sanchez.ii/PycharmProjects/PyUt/venv-pyenv-3.8.5-alternate/bin/img2py
#
from wx.lib.embeddedimage import PyEmbeddedImage

embeddedImage: PyEmbeddedImage = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAK3RFWHRDcmVhdGlvbiBUaW1l'
    b'AEZyIDE1IE5vdiAyMDAyIDEwOjA4OjI5ICswMTAw9aJixgAAAAd0SU1FB9QHBwgcNzpVudoA'
    b'AAAJcEhZcwAACvAAAArwAUKsNJgAAAAEZ0FNQQAAsY8L/GEFAAADC0lEQVR42k2T3W9TdRjH'
    b'P+ecnq7bWgp0ZejCeIkQdUCCgBoTL4yYiBlGloB3mngFFyI3Zjf8AcCFSggwL0ggMVEv8AV1'
    b'QUkgIwhcsOCFA0Y3cO26denO1m6lOz1vP5/TtYSTfE9+L8/z/J6X71fjuW90PfFWn8O0a73m'
    b'ytRLnh8wX5wbm7H58+t5Lgw6FMXME6imj9Zc2Ft43053XGjb19dpbtwEU1ko5mFxnqUnWTIj'
    b'+cKNeffLz0v8JuYVgR/6GeHP28zeWk/P5RVH+lcYdhn39jWMUgHmiiiriGkV6FwdjSdttzfm'
    b'MHXL5Un4ZpiJUewi4XSvuZ78oj/hDA0yNh1g620kXUve8AhmS/inrmAUc3TMZgzjKa9fqXL3'
    b'KcxIAEdf0Dma6O1L898oQW6caN+nRI9fxH6cJchOoL7/B3P7m6j8eL307gTpD1vYL5tuQUTX'
    b'E7F95touuD9MrCNF+88D6LE2vGPn8b69RSS1Fr45gpZ5UK86psM2jdfE+WVBIhJJJrcwPSHN'
    b'KoNpkhr5nfJEBt7aix6NoIZ+QvvutHRLx3WkPAmS1tkgziHadc9XGtY0VCuo8UdU3jlIsCot'
    b'r7nL43lR7Do78SsBNRlguSZXy0NcKYjqZaucobKAmp5i8rOTlPsHUNEWWJirz0nbLNleGkN7'
    b'dw+VqiQqASZcsk0a6LlS7a/aZAHNtnFe3YkfaJjDV4kf2IT+9+U6a4jFJZN15EpCAAduejyS'
    b'03I4BfYYdA1vbLHUblThx3OqcvyQUrtQahvK2Ynyfz2j3KPvqX9fQA0mUOcMLBPOi/PHgtUh'
    b'E82z7Xzy9ioGtqaIeFK6L/SSfuEJliTt/KKQUs5nbbwTLj/ch8fid0nwMGRi8Icwq7XGbFBl'
    b't3CzNex0SWqdEbbkBZaknV/C+srjF3EO9TAkGBFUm1qQ6ZJeA298YPJRj2JHKsI60RI5j8k7'
    b'AaPXpXxJYlLsbgvusRwoeCamhi4SDYa9IlgvSDbuwoYJWRA21Sew2BTT8wGae+kR8QbMxrnb'
    b'UGClsX4m5/8B735J6fL0cCgAAAAASUVORK5CYII=')




from typing import List
from typing import Tuple

from logging import Logger
from logging import getLogger

from deprecated import deprecated

from wx import BLACK_PEN

from wx import CANCEL
from wx import CENTRE
from wx import DC
from wx import ID_OK
from wx import OK
from wx import RED_PEN
from wx import TextEntryDialog

from pyutmodel.PyutSDMessage import PyutSDMessage

from miniogl.LineShape import LineShape
from miniogl.AnchorPoint import AnchorPoint
from miniogl.TextShape import TextShape

from ogl.sd.OglSDInstance import OglSDInstance

from ogl.OglPosition import OglPosition
from ogl.OglLink import OglLink


# TODO : Find a way to report moves from AnchorPoints to PyutSDMessage


class OglSDMessage(OglLink):
    """
    Class for a graphical message
    """
    clsLogger: Logger = getLogger(__name__)

    def __init__(self, srcSDInstance: OglSDInstance, pyutSDMessage: PyutSDMessage, dstSDInstance: OglSDInstance):
        """
        For this class use the .pyutSDMessage property to retrieve the data model

        Args:
            srcSDInstance:   Source shape OglSDInstance
            pyutSDMessage:  PyutSDMessage
            dstSDInstance:   Destination shape OglSDInstance

        """
        self._pyutSDMessage = pyutSDMessage

        super().__init__(srcShape=srcSDInstance, pyutLink=pyutSDMessage, dstShape=dstSDInstance)
        #
        # Override OglLink anchors
        #
        srcAnchor, dstAnchor = self._createAnchorPoints(srcShape=srcSDInstance, pyutSDMessage=pyutSDMessage, dstShape=dstSDInstance)
        srcAnchorPosition = srcAnchor.GetPosition()
        dstAnchorPosition = dstAnchor.GetPosition()

        self._srcAnchor: AnchorPoint = srcAnchor
        self._dstAnchor: AnchorPoint = dstAnchor

        oglSource:      OglPosition = OglPosition.tupleToOglPosition(srcAnchorPosition)
        oglDestination: OglPosition = OglPosition.tupleToOglPosition(dstAnchorPosition)
        linkLength: float = self._computeLinkLength(srcPosition=oglSource, destPosition=oglDestination)
        dx, dy            = self._computeDxDy(srcPosition=oglSource, destPosition=oglDestination)

        centerMessageX: int = round(-dy * 5 // linkLength)
        centerMessageY: int = round(dx * 5 // linkLength)

        self._messageLabel: TextShape = self.AddText(centerMessageX, centerMessageY, pyutSDMessage.message)  # font=self._defaultFont
        self.updateMessage()
        self.SetDrawArrow(True)

        assert isinstance(srcSDInstance, OglSDInstance), 'Developer Error, src of message should be an instance'
        assert isinstance(dstSDInstance, OglSDInstance), 'Developer Error, dst of message should be an instance'
        #
        # TODO:  Should I really do this?
        srcSDInstance.addLink(self)
        dstSDInstance.addLink(self)

    @property
    def pyutSDMessage(self) -> PyutSDMessage:
        """

        Returns: The pyut sd message
        """
        return self._pyutSDMessage

    def updatePositions(self):
        """
        Define the positions on lifeline (y)
        """
        src = self.sourceAnchor
        dst = self.destinationAnchor
        srcY = self._pyutSDMessage.sourceY      + src.GetParent().GetSegments()[0][1]
        dstY = self._pyutSDMessage.destinationY + dst.GetParent().GetSegments()[0][1]
        srcX = 0
        dstX = 0

        src.SetPosition(srcX, srcY)
        dst.SetPosition(dstX, dstY)

    def updateMessage(self):
        """
        Update the message
        """
        text:      str       = self._pyutSDMessage.message
        textShape: TextShape = self._messageLabel
        # Don't draw blank messages
        if text.strip() != "":
            textShape.text = text
            textShape.SetVisible(True)
        else:
            textShape.SetVisible(False)

    @deprecated(reason='Use the .pyutSDMessage property')
    def getPyutObject(self) -> PyutSDMessage:
        """
        override

        Returns: my pyut sd message
        """
        return self._pyutSDMessage

    def Draw(self, dc: DC,  withChildren: bool = False):
        """
        Called for drawing the contents of links.

        Args:
            dc:     Device context
            withChildren:   `True` draw the children
        """
        self.updateMessage()

        srcAnchor, dstAnchor = self.getAnchors()

        srcX, srcY = srcAnchor.GetPosition()
        dstX, dstY = dstAnchor.GetPosition()

        if self._selected is True:
            dc.SetPen(RED_PEN)

        dc.DrawLine(srcX, srcY, dstX, dstY)
        self.DrawArrow(dc, srcAnchor.GetPosition(), dstAnchor.GetPosition())
        self.DrawChildren(dc=dc)

        dc.SetPen(BLACK_PEN)

    def Detach(self):
        """
        Override OglLink because are ends are OglSDInstance's
        """
        if self._diagram is not None and not self._protected:
            LineShape.Detach(self)
            self._srcAnchor.SetProtected(False)
            self._dstAnchor.SetProtected(False)
            self._srcAnchor.Detach()
            self._dstAnchor.Detach()

            self._detachFromOglEnds()
            # TODO:
            # pyutSDMessage: PyutSDMessage = self.pyutSDMessage
            # I don't think anything needs to be done because once
            # the Ogl instance is gone the model should disappear

    def _detachFromOglEnds(self):
        """
        Override base class
        """
        src: OglSDInstance = self.sourceShape
        dst: OglSDInstance = self.destinationShape
        assert isinstance(src, OglSDInstance), 'Developer Error, src of message should be an instance'
        assert isinstance(dst, OglSDInstance), 'Developer Error, dst of message should be an instance'

        links: List[OglSDMessage] = src.links
        links.remove(self)

        links = dst.links
        links.remove(self)

    def OnLeftDClick(self, event):
        """
        Callback for left double clicks.

        """
        dlg = TextEntryDialog(None, "Message", "Enter message name", self._pyutSDMessage.getMessage(), OK | CANCEL | CENTRE)
        if dlg.ShowModal() == ID_OK:
            self._pyutSDMessage.setMessage(dlg.GetValue())
        dlg.Destroy()

    def _createAnchorPoints(self, srcShape: OglSDInstance, dstShape: OglSDInstance,
                            pyutSDMessage: PyutSDMessage) -> Tuple[AnchorPoint, AnchorPoint]:
        """
        Royal ready to heat rice

        Args:
            dstShape:
            srcShape:
            pyutSDMessage

        Returns:  A tuple of anchor points for the source and destination shapes
        """
        srcY: int = pyutSDMessage.sourceY      - srcShape.lifeline.GetPosition()[1]
        dstY: int = pyutSDMessage.destinationY - dstShape.lifeline.GetPosition()[1]

        srcAnchor: AnchorPoint = srcShape.lifeline.AddAnchor(0, srcY)
        dstAnchor: AnchorPoint = dstShape.lifeline.AddAnchor(0, dstY)
        srcAnchor.SetStayOnBorder(False)
        dstAnchor.SetStayOnBorder(False)
        srcAnchor.SetStayInside(True)
        dstAnchor.SetStayInside(True)
        srcAnchor.SetVisible(True)
        dstAnchor.SetVisible(True)
        srcAnchor.draggable = True
        dstAnchor.draggable = True

        return srcAnchor, dstAnchor

    def __repr__(self) -> str:
        msg: str = self._pyutSDMessage.message
        return f'OglSDMessage[id: {self._id} {msg=}]'




from wx import PENSTYLE_LONG_DASH

from wx import Pen

from miniogl.Shape import Shape

from pyutmodel.PyutLink import PyutLink
from ogl.OglLink import OglLink
from ogl.OglObject import OglObject


class OglNoteLink(OglLink):
    """
    A note like link, with dashed line and no arrows.
    To get a new link, you should use the `OglLinkFactory` and specify
    the kind of link you want, OGL_NOTELINK for an instance of this class.

    """

    def __init__(self, srcShape: OglObject, pyutLink: PyutLink, dstShape: OglObject):
        """

        Args:
            srcShape:  Source shape
            pyutLink:  Conceptual links associated with the graphical links.
            dstShape: Destination shape
        """
        super().__init__(srcShape, pyutLink, dstShape)
        self.SetDrawArrow(False)
        self.SetPen(Pen("BLACK", 1, PENSTYLE_LONG_DASH))

    def __repr__(self):

        srcShape:  Shape = self.sourceShape
        destShape: Shape = self.destinationShape
        sourceId:  int   = srcShape.id
        destId:    int   = destShape.id
        return f'OglNoteLink - from: id: {sourceId} {self.sourceShape}  to: id: {destId} {self.destinationShape}'



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()
LICENSE = (HERE / 'LICENSE').read_text()

setup(
    name="ogl",
    version="0.70.25",
    author='Humberto A. Sanchez II',
    author_email='Humberto.A.Sanchez.II@gmail.com',
    maintainer='Humberto A. Sanchez II',
    maintainer_email='humberto.a.sanchez.ii@gmail.com',
    description='External Pyut Graphical Shapes',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/hasii2011/ogl",
    package_data={
        'miniogl':              ['py.typed'],
        'ogl':                  ['py.typed'],
        'ogl.events':           ['py.typed'],
        'ogl.preferences':      ['py.typed'],
        'ogl.resources':        ['py.typed'],
        'ogl.resources.img':             ['py.typed'],
        'ogl.resources.img.textdetails': ['py.typed'],
        'ogl.sd':               ['py.typed'],
        'ogl.ui':               ['py.typed'],
        'ogl.ui.valuecontrols': ['py.typed'],
    },

    packages=[
        'miniogl',
        'ogl',
        'ogl.events',
        'ogl.preferences',
        'ogl.resources', 'ogl.resources.img', 'ogl.resources.img.textdetails',
        'ogl.sd',
        'ogl.ui', 'ogl.ui.valuecontrols',
    ],
    install_requires=['Deprecated~=1.2.13', 'pyutmodel==1.4.3', 'hasiihelper==0.2.0', 'hasiicommon==0.2.2', 'wxPython~=4.2.0'],
)




from typing import Tuple
from typing import cast

from deprecated import deprecated

from wx import DC

from miniogl.Shape import Shape
from miniogl.MiniOglUtils import sign
from miniogl.SizerShape import SizerShape
from miniogl.RectangleShapeModel import RectangleShapeModel


class RectangleShape(Shape):
    """
    A rectangle shape.
    """
    def __init__(self, x: int = 0, y: int = 0, width: int = 0, height: int = 0, parent=None):
        """

        Args:
            x:      Shape's x position
            y:      Shape's y position
            width:  Shape's width
            height: Shape's height
            parent: The shape's parent, if any
        """
        super().__init__(x, y, parent)

        self._width:  int = width   # width and height can be < 0 !!!
        self._height: int = height

        self._drawFrame: bool = True
        self._resizable: bool = True

        self._topLeftSizer:  SizerShape = cast(SizerShape, None)
        self._topRightSizer: SizerShape = cast(SizerShape, None)
        self._botLeftSizer:  SizerShape = cast(SizerShape, None)
        self._botRightSizer: SizerShape = cast(SizerShape, None)

        self._ox: int = 0   # This is done in Shape but Pycharm can't see this in the ShowSizer() code
        # set the model of the shape (MVC pattern)
        self._model: RectangleShapeModel = RectangleShapeModel(self)
        self._model.SetPosition(x=x, y=y)
        self._model.SetSize(width=width, height=height)

    @property
    def selected(self) -> bool:
        """
        Override Shape

        Returns: 'True' if selected 'False' otherwise
        """
        return self._selected

    @selected.setter
    def selected(self, state: bool):
        self._selected = state
        if self._resizable:
            self.ShowSizers(state)

    def SetResizable(self, state: bool):
        """
        Set the resizable flag.

        @param state
        """
        self._resizable = state

    def GetResizable(self) -> bool:
        """
        Get the resizable flag.

        @return bool
        """
        return self._resizable

    def GetTopLeft(self) -> Tuple[int, int]:
        """
        Get the coordinates of the top left point in diagram coordinates.

        @return (double, double)
        """
        x, y = self.GetPosition()
        x -= self._ox
        y -= self._oy
        width, height = self.GetSize()
        if width < 0:
            x += width
        if height < 0:
            y += height
        return x, y

    def SetTopLeft(self, x, y):
        """
        Set the position of the top left point.

        @param  x
        @param y : new position
        """
        x += self._ox
        y += self._oy
        width, height = self.GetSize()
        if width < 0:
            x -= width
        if height < 0:
            y -= height
        self._x, self._y = x, y

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Draw the rectangle on the dc.

        Args:
            dc:
            withChildren:

        Returns:

        """
        if self._visible:
            Shape.Draw(self, dc, False)
            if self._drawFrame:
                sx, sy = self.GetPosition()
                sx, sy = sx - self._ox, sy - self._oy
                width, height = self.GetSize()

                dc.DrawRectangle(sx, sy, width, height)
            if withChildren:
                self.DrawChildren(dc)
            if self._topLeftSizer is not None:
                self._topLeftSizer.Draw(dc, False)

    def DrawBorder(self, dc):
        """
        Draw the border of the shape, for fast rendering.
        """
        Shape.DrawBorder(self, dc)
        sx, sy = self.GetPosition()
        sx, sy = sx - self._ox, sy - self._oy
        width, height = self.GetSize()
        dc.DrawRectangle(sx, sy, width, height)

    def SetDrawFrame(self, draw):
        """
        Choose to draw a frame around the rectangle.

        @param bool draw : True to draw the frame
        """
        self._drawFrame = draw

    def Inside(self, x, y) -> bool:
        """
        True if (x, y) is inside the rectangle.

        @param  x
        @param y
        @return bool
        """
        # this also works if width and/or height is negative.
        sx, sy = self.GetPosition()
        # take a minimum of 4 pixels for the selection
        width, height = self.GetSize()
        width  = sign(width)  * max(abs(width),  4)
        height = sign(height) * max(abs(height), 4)
        topLeftX: int = sx - self._ox
        topLeftY: int = sy - self._oy

        topXLeftIsInside:  bool  = x > topLeftX
        topXRightIsInside: bool  = x < topLeftX + width
        topYLeftIsInside:  bool  = y > topLeftY
        topYRightIsInside: bool  = y < topLeftY + height
        if topXLeftIsInside and topXRightIsInside and topYLeftIsInside and topYRightIsInside:
            return True
        else:
            return False

    def GetSize(self) -> Tuple[int, int]:
        """
        Get the size of the rectangle.

        Returns:
            A tuple of width, height
        """
        return self._width, self._height

    def GetHeight(self):
        return self._height

    @deprecated(reason='Use the .selected property')
    def SetSelected(self, state: bool = True):
        """
        Select the shape.

        @param state
        """
        Shape.SetSelected(self, state)
        if self._resizable:
            self.ShowSizers(state)

    def Detach(self):
        """
        Detach the shape from its diagram.
        This is the way to delete a shape. All anchor points are also
        removed, and link lines too.
        """
        # do not detach a protected shape
        if self._diagram is not None and not self._protected:
            Shape.Detach(self)
            self.ShowSizers(False)

    def ShowSizers(self, state: bool = True):
        """
        Show the four sizer shapes if state is True.

        @param state
        """
        width, height = self.GetSize()
        if state and not self._topLeftSizer:
            self._topLeftSizer  = SizerShape(-self._ox, -self._oy, self)
            self._topRightSizer = SizerShape(-self._ox + width - 1, self._oy, self)
            self._botLeftSizer  = SizerShape(-self._ox, -self._oy + height - 1, self)
            self._botRightSizer = SizerShape(-self._ox + width - 1, -self._oy + height - 1, self)

            self._diagram.AddShape(self._topLeftSizer)
            self._diagram.AddShape(self._topRightSizer)
            self._diagram.AddShape(self._botLeftSizer)
            self._diagram.AddShape(self._botRightSizer)
        elif not state and self._topLeftSizer is not None:
            self._topLeftSizer.Detach()
            self._topRightSizer.Detach()
            self._botLeftSizer.Detach()
            self._botRightSizer.Detach()

            self._topLeftSizer  = cast(SizerShape, None)
            self._topRightSizer = cast(SizerShape, None)
            self._botLeftSizer  = cast(SizerShape, None)
            self._botRightSizer = cast(SizerShape, None)

    def SetSize(self, width, height):
        """
        Set the size of the rectangle.

        @param  width
        @param height
        """
        self._width, self._height = width, height

        if self.HasDiagramFrame():
            self.UpdateModel()

        for anchor in self._anchors:
            ax, ay = anchor.GetPosition()
            # Reset position to stick the border
            anchor.SetPosition(ax, ay)

    def Resize(self, sizer, x, y):
        """
        Resize the rectangle according to the new position of the sizer.
        Not used to programmatically resize a shape. Use `SetSize` for this.

        Args:
            sizer:
            x:      x position of the sizer
            y:      y position of the sizer

        """
        tlx, tly = self.GetTopLeft()
        w, h = self.GetSize()
        sw, sh = sign(w), sign(h)
        w, h = abs(w), abs(h)
        if sizer is self._topLeftSizer:
            nw = sw * (w - x + tlx)
            nh = sh * (h - y + tly)
            self._ox = self._ox * nw // w
            self._oy = self._oy * nh // h
            self.SetSize(nw, nh)
            self.SetTopLeft(x, y)
            self._topRightSizer.SetRelativePosition(nw - 1, 0)
            self._botLeftSizer.SetRelativePosition(0, nh - 1)
            self._botRightSizer.SetRelativePosition(nw - 1, nh - 1)
        elif sizer is self._topRightSizer:
            nw = sw * (x - tlx)
            nh = sh * (tly + h - y)
            self.SetTopLeft(tlx, y)
            self.SetSize(nw + 1, nh)
            self._topRightSizer.SetRelativePosition(nw, 0)
            self._botRightSizer.SetRelativePosition(nw, nh - 1)
            self._botLeftSizer.SetRelativePosition(0, nh - 1)
        elif sizer is self._botLeftSizer:
            nw = sw * (w - x + tlx)
            nh = sh * (y - tly)
            self.SetTopLeft(x, tly)
            self.SetSize(nw, nh + 1)
            self._botLeftSizer.SetRelativePosition(0, nh)
            self._botRightSizer.SetRelativePosition(nw - 1, nh)
            self._topRightSizer.SetRelativePosition(nw - 1, 0)
        elif sizer is self._botRightSizer:
            nw = sw * (x - tlx)
            nh = sh * (y - tly)
            self.SetSize(nw + 1, nh + 1)
            self._botLeftSizer.SetRelativePosition(0, nh)
            self._botRightSizer.SetRelativePosition(nw, nh)
            self._topRightSizer.SetRelativePosition(nw, 0)

    def UpdateFromModel(self):
        """
        Updates the shape position and size from the model in the light of a
        change of state of the diagram frame (here it's only for the zoom)
        """

        #  change the position of the shape from the model
        Shape.UpdateFromModel(self)

        #  get the model size
        width, height = self.GetModel().GetSize()

        #  get the diagram frame ratio between the shape and the model
        ratio = self.GetDiagram().GetPanel().currentZoom

        # set the new size to the shape.
        self._width  = round(width * ratio)
        self._height = round(height * ratio)

    def UpdateModel(self):
        """
        Updates the model when the shape (view) is moved or resized.
        """
        #  change the coordinates of model
        Shape.UpdateModel(self)

        #  get the size of the shape (view)
        width, height = self.GetSize()

        # get the ratio between the model and the shape (view) from
        # the diagram frame where the shape is displayed.
        ratio = self.GetDiagram().GetPanel().currentZoom

        # set the new size to the model.
        self.GetModel().SetSize(round(width//ratio), round(height//ratio))
